/*
 * EpsilonRemoverIncoming.h
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#ifndef EPSILON_REMOVER_INCOMING_H_
#define EPSILON_REMOVER_INCOMING_H_

#include <core/multipleDispatch.hpp>
#include <automaton/Automaton.h>

#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include "../properties/EpsilonClosure.h"

namespace automaton {

namespace simplify {

class EpsilonRemoverIncoming : public std::SingleDispatch<EpsilonRemoverIncoming, automaton::Automaton, const automaton::AutomatonBase &> {
public:
	static automaton::Automaton remove( const automaton::Automaton & automaton );

	/**
	 * Computes epsilon closure of a state in epsilon nonfree automaton
	 */
	template < class SymbolType, class EpsilonType, class StateType >
	static automaton::NFA < SymbolType, StateType > remove( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm );
	template < class SymbolType, class StateType >
	static automaton::MultiInitialStateNFA < SymbolType, StateType > remove( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm );
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > remove( const automaton::NFA < SymbolType, StateType > & fsm );
	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, StateType > remove( const automaton::DFA < SymbolType, StateType > & fsm );

};

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, StateType > EpsilonRemoverIncoming::remove(const automaton::DFA < SymbolType, StateType > & origFSM) {
	return origFSM;
}

template < class SymbolType, class StateType >
automaton::MultiInitialStateNFA < SymbolType, StateType > EpsilonRemoverIncoming::remove(const automaton::MultiInitialStateNFA < SymbolType, StateType > & origFSM) {
	return origFSM;
}

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > EpsilonRemoverIncoming::remove(const automaton::NFA < SymbolType, StateType > & origFSM) {
	return origFSM;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::NFA < SymbolType, StateType > EpsilonRemoverIncoming::remove( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & origFSM ) {
	automaton::NFA < SymbolType, StateType > fsm(origFSM.getInitialState());

	fsm.setStates( origFSM.getStates() );
	fsm.setInputAlphabet( origFSM.getInputAlphabet() );

	std::map<std::pair<StateType, SymbolType>, std::set<StateType>> origTransitions = origFSM.getSymbolTransitions();

	/**
	 * Step 1 from Melichar 2.41
	 */
	for( const auto & from : origFSM.getStates( ) )
		for( const auto & fromClosure : automaton::properties::EpsilonClosure::epsilonClosure( origFSM, from ) )
			for( const auto & symbol : origFSM.getInputAlphabet() ) {
				auto transitions = origTransitions.find(std::make_pair(fromClosure, symbol));
				if(transitions == origTransitions.end()) continue;

				for( const auto & to : transitions->second )
					fsm.addTransition( from, symbol, to );
			}

	/**
	 * Step 2 from Melichar 2.41
	 */
	const std::set<StateType> & F = origFSM.getFinalStates( );
	for( const auto & q : fsm.getStates( ) ) {
		const std::set<StateType> & cl = automaton::properties::EpsilonClosure::epsilonClosure( origFSM, q );
		std::set<StateType> intersect;

		set_intersection( cl.begin(), cl.end(), F.begin(), F.end(), std::inserter( intersect, intersect.begin() ) );
		if( intersect.size( ) != 0 )
			fsm.addFinalState( q );
	}

	return fsm;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* EPSILON_REMOVER_INCOMING_H_ */
