/*
 * Normalize.cpp
 *
 *  Created on: Dec 9, 2013
 *	Author: Jan Travnicek
 */

#include "Normalize.h"

#include <map>
#include <deque>
#include <vector>
#include <algorithm>
#include "exception/CommonException.h"

#include "alphabet/Symbol.h"

#include "automaton/Automaton.h"

namespace automaton {

namespace simplify {

automaton::Automaton Normalize::normalize(const automaton::Automaton& automaton) {
	return dispatch(automaton.getData());
}

automaton::DFA<> Normalize::normalize(const automaton::DFA<>& fsm) {
	int counter = 0;
	std::map<DefaultStateType, int > normalizationData;
	std::deque<DefaultStateType > processingData;

	normalizationData.insert(std::make_pair(fsm.getInitialState(), counter++));
	processingData.push_back(fsm.getInitialState());

	while(!processingData.empty()) {
		DefaultStateType current = processingData.front();
		processingData.pop_front();

		std::map<std::pair<DefaultStateType, DefaultSymbolType>, DefaultStateType> transitionsMap = fsm.getTransitionsFromState(current);
		// Transitions are trivialy sorted by input symbol (from state is the same)

		for(auto iter = transitionsMap.begin(); iter != transitionsMap.end(); iter++) {
			if(normalizationData.find(iter->second) == normalizationData.end()) {
				normalizationData.insert(std::pair<DefaultStateType, int>(iter->second, counter++));
				processingData.push_back(iter->second);
			}
		}
	}

	if(normalizationData.size() != fsm.getStates().size()) {
		throw exception::CommonException("Automaton normalize require minimal deterministic finite automaton");
	}

	automaton::DFA<> result(DefaultStateType(normalizationData.find(fsm.getInitialState())->second));

	result.setInputAlphabet(fsm.getInputAlphabet());

	for(auto iter = fsm.getStates().begin(); iter != fsm.getStates().end(); iter++) {
		result.addState(DefaultStateType(normalizationData.find(*iter)->second));
	}

	for(auto iter = fsm.getFinalStates().begin(); iter != fsm.getFinalStates().end(); iter++) {
		result.addFinalState(DefaultStateType(normalizationData.find(*iter)->second));
	}

	for(auto iter = fsm.getTransitions().begin(); iter != fsm.getTransitions().end(); iter++) {
		result.addTransition(DefaultStateType(normalizationData.find(iter->first.first)->second), iter->first.second, DefaultStateType(normalizationData.find(iter->second)->second));
	}

	return result;
}

auto NormalizeDFA = Normalize::RegistratorWrapper<automaton::DFA<>, automaton::DFA<>>(Normalize::normalize);

automaton::DPDA < > Normalize::normalize(const automaton::DPDA < > & pda) {
	int counterState = 0;
	std::map<DefaultStateType, int > normalizationDataState;
	int counterSymbol = 0;
	std::map<DefaultSymbolType, int > normalizationDataSymbol;
	std::deque< DefaultStateType > processingData;

	normalizationDataState.insert(std::make_pair(pda.getInitialState(), counterState++));
	normalizationDataSymbol.insert(std::make_pair(pda.getInitialSymbol(), counterSymbol++));
	processingData.push_back(pda.getInitialState());

	while(!processingData.empty()) {
		DefaultStateType current = processingData.front();
		processingData.pop_front();

		std::map<std::pair<std::variant<string::Epsilon < >, DefaultSymbolType>, std::vector<DefaultSymbolType> >, std::pair<DefaultStateType, std::vector<DefaultSymbolType> > > transform;
		bool stateFinished = true;
		// For each transition from state current
		for(const std::pair<const std::tuple<DefaultStateType, std::variant<string::Epsilon < >, DefaultSymbolType>, std::vector<DefaultSymbolType> >, std::pair<DefaultStateType, std::vector<DefaultSymbolType> > >& iter : pda.getTransitionsFromState(current)) {
			// look whether all poped symbols are already transformed
			if(std::all_of(std::get<2>(iter.first).begin(), std::get<2>(iter.first).end(), [&](const DefaultSymbolType& symbol) { return normalizationDataSymbol.find(symbol) != normalizationDataSymbol.end(); })) {
				std::vector<DefaultSymbolType> transformedSymbols;
				// if so compute vector of transformed poped symbol -- this can be compared to other vectors of transformed symbols and the order can by trusted
				for(const DefaultSymbolType& symbol : std::get<2>(iter.first)) {
					transformedSymbols.push_back(DefaultSymbolType(normalizationDataSymbol.find(symbol)->second));
				}

				// then first order is by input symbol and the second is this vector transfomed symbols 
				transform.insert(std::make_pair(std::make_pair(std::get<1>(iter.first), transformedSymbols), iter.second));
			} else {
				stateFinished = false;
			}
		}
		// if there was a single transition with undefined transformation for poped symbol process this state again later
		if(!stateFinished) {
			processingData.push_back(current);
		}
		// now transitions are trivialy sorted by input symbol and the already transformed symbols

		// for each pair state, pushed symbols in order given by input symbol and transformed poped symbols
		for(const auto& iter : transform) {
			const auto& iters = iter.second;
			// if the state is new asign a unique number to it and mark it for processing
			if( normalizationDataState.find(iters.first) == normalizationDataState.end() ) {
				normalizationDataState.insert(std::make_pair(iters.first, counterState++));
				processingData.push_back(iters.first);
			}
			// if the symbols in order given by order of pushing are new asign a unique number to them
			for(const DefaultSymbolType& iter2 : iters.second) {
				if( normalizationDataSymbol.find(iter2) == normalizationDataSymbol.end() ) {
					normalizationDataSymbol.insert(std::make_pair(iter2, counterSymbol++));
				}
			}
		}
	}

	if(normalizationDataState.size() != pda.getStates().size() || normalizationDataSymbol.size() != pda.getPushdownStoreAlphabet().size()) {
		throw exception::CommonException("Automaton normalize require connected deterministic pushdown automaton");
	}

	automaton::DPDA < > result(DefaultStateType(normalizationDataState.find(pda.getInitialState())->second), DefaultSymbolType(normalizationDataSymbol.find(pda.getInitialSymbol())->second));
	result.setInputAlphabet(pda.getInputAlphabet());

	for(const DefaultSymbolType & symbol : pda.getPushdownStoreAlphabet())
		result.addPushdownStoreSymbol(DefaultSymbolType(normalizationDataSymbol.find(symbol)->second));

	for(const DefaultStateType & state : pda.getStates())
		result.addState(DefaultStateType(normalizationDataState.find(state)->second));

	for(const DefaultStateType & state : pda.getFinalStates())
		result.addFinalState(DefaultStateType(normalizationDataState.find(state)->second));

	for(const auto & transition : pda.getTransitions()) {
		std::vector<DefaultSymbolType> pop;
		for(const auto& elem : std::get<2>(transition.first)) {
			pop.push_back(DefaultSymbolType(normalizationDataSymbol.find(elem)->second));
		}
		std::vector<DefaultSymbolType> push;
		for(const auto& elem : transition.second.second) {
			push.push_back(DefaultSymbolType(normalizationDataSymbol.find(elem)->second));
		}
		result.addTransition(DefaultStateType(normalizationDataState.find(std::get<0>(transition.first))->second), std::get<1>(transition.first), pop, DefaultStateType(normalizationDataState.find(transition.second.first)->second), push);
	}

	return result;
}

auto NormalizeDPDA = Normalize::RegistratorWrapper<automaton::DPDA < >, automaton::DPDA < > >(Normalize::normalize);

} /* namespace simplify */

} /* namespace automaton */
