/*
 * XmlApi.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef XML_API_HPP_
#define XML_API_HPP_

#include <functional>
#include <deque>
#include <vector>
#include <map>
#include <string>
#include "sax/Token.h"
#include "sax/FromXMLParserHelper.h"

#include "object/ObjectBase.h"
#include "base/WrapperBase.hpp"
#include "exception/CommonException.h"

#include <global/GlobalData.h>

namespace alib {

template < typename T, typename Enable = void >
struct xmlApi { };

class xmlApiInputContext : public std::deque < sax::Token >::iterator {
	template < typename T, typename Enable >
	friend struct xmlApi;

	std::map < std::string, std::map < int, void * > > idToInstanceContexts;
	int idMaxContext;

public:
	xmlApiInputContext ( std::deque < sax::Token >::iterator iter ) : std::deque < sax::Token >::iterator ( iter ), idMaxContext ( 0 ) {
	}

	~xmlApiInputContext ( ) {
		for ( const std::pair < const std::string, std::map < int, void * > > context : idToInstanceContexts ) {
			std::function < void ( void * ) > & deleteCallback = deleteCallbacks ( ).find ( context.first )->second;

			for ( const std::pair < const int, void * > entry : context.second )
				deleteCallback ( entry.second );
		}
	}

	std::map < int, void * > & idToInstance ( const std::string & name ) {
		return idToInstanceContexts[name];
	}

	int & idMax ( ) {
		return idMaxContext;
	}

	static std::map < std::string, std::function < void ( void * ) > > & deleteCallbacks ( ) {
		static std::map < std::string, std::function < void ( void * ) > > res;

		return res;
	}

	const std::map < std::string, std::map < int, void * > > & dump ( ) const {
		return idToInstanceContexts;
	}

};

class xmlApiOutputContext : public std::deque < sax::Token > {
	template < typename T, typename Enable >
	friend struct xmlApi;

	std::map < std::string, std::map < void *, int > > instanceToIdContexts;
	int idMaxContext;

public:
	xmlApiOutputContext ( ) : idMaxContext ( 0 ) {
	}

	std::map < void *, int > & instanceToId ( const std::string & name ) {
		return instanceToIdContexts[name];
	}

	int & idMax ( ) {
		return idMaxContext;
	}

};

template < typename Group >
struct xmlApi < Group, typename std::enable_if < std::is_base_of < WrapperBaseBase, Group >::value >::type > {
	class ParserRegisterBase {
	public:
		virtual Group parse ( std::deque < sax::Token >::iterator & input ) = 0;
	};

	static std::map < std::string, typename xmlApi < Group >::ParserRegisterBase * > & parseFunctions ( ) {
		static std::map < std::string, typename xmlApi < Group >::ParserRegisterBase * > res;

		return res;
	}

public:
	class InputContextDeleter {
	public:
		InputContextDeleter ( ) {
			std::function < void ( void * ) > deleteCallback = [] ( void * pointer ) { delete ( Group * ) pointer; };
			xmlApiInputContext::deleteCallbacks ( ).insert ( std::make_pair ( std::type_name < Group > ( ), deleteCallback ) );
		}

	};

	template < class Type >
	class ParserRegister : public ParserRegisterBase {
		std::function < Type ( std::deque < sax::Token >::iterator & ) > parseFunction;

	public:
		ParserRegister ( ) : parseFunction ( Type::parse ) {
			parseFunctions ( ).insert ( std::make_pair ( Type::getXmlTagName(), this ) );
		}

		virtual Group parse ( std::deque < sax::Token >::iterator & input ) {
			return Group ( parseFunction ( input ) );
		}

	};

	static Group parse ( std::deque < sax::Token >::iterator & data ) {
		xmlApiInputContext & input = ( xmlApiInputContext & ) data;
		sax::FromXMLParserHelper::skipAttributes ( input, sax::Token::TokenType::START_ELEMENT );

		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );
		if ( /* common::GlobalData::optimizeXml && */ tagName == "Ref" ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Ref" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ATTRIBUTE, "id" );
			int id = std::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( input, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ATTRIBUTE, "id" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Ref" );
			std::map < int, void * >::iterator elem = input.idToInstance ( std::type_name < Group > ( ) ).find ( id );

			if ( elem == input.idToInstance ( std::type_name < Group > ( ) ).end ( ) ) {
				std::cerr << input.dump ( ) << std::endl;
				throw exception::CommonException ( "XML Inconsistent ( id not found " + std::to_string ( id  ) + " )" );
			}

			return * ( ( Group * ) elem->second );
		} else {
			typename std::map < std::string, ParserRegisterBase * >::iterator callback = parseFunctions ( ).find ( tagName );

			if ( callback == parseFunctions ( ).end ( ) ) throw exception::CommonException ( "Parse callback for " + tagName + " tag not registered." );

			/* find out if this is a base for reference */
			bool ref = sax::FromXMLParserHelper::isToken ( input + 1, sax::Token::TokenType::START_ATTRIBUTE, "ref" );
			int id = 0;
			if ( ref ) {
				std::deque < sax::Token >::iterator tmp = input + 2;
				id = std::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( tmp, sax::Token::TokenType::CHARACTER ) );
			}

			Group * res = new Group ( callback->second->parse ( input ) );

			/* if object is a base of reference, register it */
			if ( ref )
				input.idToInstance ( std::type_name < Group > ( ) ).insert ( std::make_pair ( id, ( void * ) res ) );

			return * res;
		}
	}

	static bool first ( const std::deque < sax::Token >::const_iterator & input ) {
		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		if ( tagName == "Ref" ) return true;

		typename std::map < std::string, ParserRegisterBase * >::iterator callback = parseFunctions ( ).find ( tagName );

		if ( callback == parseFunctions ( ).end ( ) ) return false;

		return true;
	}

	static void compose ( std::deque < sax::Token > & output, const Group & data ) {
		xmlApiOutputContext & context = ( xmlApiOutputContext & ) output;

		typename std::map < void *, int >::iterator elem = context.instanceToId ( std::type_name < Group > ( ) ).find ( const_cast< void * >(static_cast< const void * > ( & data.getData ( ) ) ) );

		if ( common::GlobalData::optimizeXml && elem != context.instanceToId ( std::type_name < Group > ( ) ).end ( ) ) {
			output.emplace_back ( "Ref", sax::Token::TokenType::START_ELEMENT );
			output.emplace_back ( "id", sax::Token::TokenType::START_ATTRIBUTE );
			output.emplace_back ( std::to_string ( elem->second ), sax::Token::TokenType::CHARACTER );
			output.emplace_back ( "id", sax::Token::TokenType::END_ATTRIBUTE );
			output.emplace_back ( "Ref", sax::Token::TokenType::END_ELEMENT );
		} else {
			/* get current position in token stream */
			size_t pos = output.size();
			int id = 0;
			if ( common::GlobalData::optimizeXml ) {
				id = context.idMax ( )++;
			}

			/* encode referenced object */
			data.getData ( ).compose ( output );

			/* emplace referenced attribute */
			if ( common::GlobalData::optimizeXml ) {
				output.emplace ( output.begin ( ) + pos + 1, "ref", sax::Token::TokenType::START_ATTRIBUTE );
				output.emplace ( output.begin ( ) + pos + 2, std::to_string ( id ) , sax::Token::TokenType::CHARACTER );
				output.emplace ( output.begin ( ) + pos + 3, "ref", sax::Token::TokenType::END_ATTRIBUTE );

				context.instanceToId ( std::type_name < Group > ( ) ).insert ( std::make_pair ( const_cast< void * > ( static_cast< const void * > ( & data.getData ( ) ) ), id ) );
			}
		}
	}

};

template < typename Type >
struct xmlApi < Type, typename std::enable_if < std::is_base_of < ObjectBase, Type >::value >::type > {
	static Type parse ( std::deque < sax::Token >::iterator & input ) {
		return Type::parse ( input );
	}

	static bool first ( const std::deque < sax::Token >::const_iterator & input ) {
		return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, Type::getXmlTagName() );
	}

	static void compose ( std::deque < sax::Token > & output, const Type & data ) {
		data.compose ( output );
	}

};

} /* namespace alib */

#endif /* XML_API_HPP_ */
