/*
 * NPDTA.h
 *
 *  Created on: 10. 5. 2016
 *      Author: Jakub Doupal
 */

#ifndef NPDTA_H_
#define NPDTA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class InputAlphabet;
class OutputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Translation Automaton
 */
template < class SymbolType, class EpsilonType, class StateType >
class NPDTA : public AutomatonBase, public std::Components < NPDTA < SymbolType, EpsilonType, StateType >, SymbolType, std::tuple < InputAlphabet, OutputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::set < std::tuple < StateType, std::vector < SymbolType >, std::vector < SymbolType > > > > transitions;

public:
	explicit NPDTA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > outputAlphabet, std::set < SymbolType > pushdownStoreSymbol, StateType initialState, SymbolType initialPushdownSymbol, std::set < StateType > finalStates );
	explicit NPDTA ( StateType initialState, SymbolType initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;


	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	const std::set < SymbolType > & getOutputAlphabet ( ) const {
		return this->template accessComponent < OutputAlphabet > ( ).get ( );
	}

	bool addOutputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < OutputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addOutputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < OutputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setOutputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < OutputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeOutputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < OutputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the NPDTA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the NPDTA (state, input symbol, stack symbol, output symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, std::variant < EpsilonType, SymbolType > input, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push, std::vector < SymbolType > output );

	bool addTransition ( StateType from, SymbolType input, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push, std::vector < SymbolType > output );

	bool addTransition ( StateType from, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push, std::vector < SymbolType > output );

	/**
	 * Removes the transition from the NPDTA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the NPDTA
	 */
	bool removeTransition ( const StateType & from, const std::variant < EpsilonType, SymbolType > & input, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push, const std::vector < SymbolType > & output );

	bool removeTransition ( const StateType & from, const SymbolType & input, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push, const std::vector < SymbolType > & output );

	bool removeTransition ( const StateType & from, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push, const std::vector < SymbolType > & output );

	/**
	 * @return NPDTA transitions
	 */
	const std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::set < std::tuple < StateType, std::vector < SymbolType >, std::vector < SymbolType > > > > & getTransitions ( ) const;

	std::set < std::pair < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::tuple < StateType, std::vector < SymbolType >, std::vector < SymbolType > > > > getTransitionsFromState ( const StateType & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NPDTA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "NPDTA";

		return xmlTagName;
	}

	static NPDTA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, NPDTA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class EpsilonType, class StateType >
NPDTA < SymbolType, EpsilonType, StateType >::NPDTA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > outputAlphabet, std::set < SymbolType > pushdownStoreAlphabet, StateType initialState, SymbolType initialSymbol, std::set < StateType > finalStates ) : std::Components < NPDTA, SymbolType, std::tuple < InputAlphabet, OutputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( outputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class SymbolType, class EpsilonType, class StateType >
NPDTA < SymbolType, EpsilonType, StateType >::NPDTA(StateType initialState, SymbolType initialPushdownSymbol) : NPDTA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, std::set < SymbolType > { }, std::set < SymbolType > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* NPDTA < SymbolType, EpsilonType, StateType >::clone() const {
	return new NPDTA(*this);
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* NPDTA < SymbolType, EpsilonType, StateType >::plunder() && {
	return new NPDTA(std::move(*this));
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDTA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::variant<EpsilonType, SymbolType> input, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push, std::vector<SymbolType> output) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is<SymbolType>() && !getInputAlphabet().count(input.template get<SymbolType>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	for(const SymbolType& popSymbol : pop) {
		if (!getPushdownStoreAlphabet().count(popSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( popSymbol ) + "\" doesn't exist.");
		}
	}

	for(const SymbolType& pushSymbol : push) {
		if (!getPushdownStoreAlphabet().count(pushSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");
		}
	}

	for(const SymbolType& outputSymbol : output) {
		if (!getOutputAlphabet().count(outputSymbol)) {
			throw AutomatonException("Output symbol \"" + std::to_string ( outputSymbol ) + "\" doesn't exist.");
		}
	}

	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> > key(std::move(from), std::move(input), std::move(pop));
	std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > value(std::move(to), std::move(push), std::move(output));

	return transitions[std::move(key)].insert(std::move(value)).second;
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDTA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, SymbolType input, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push, std::vector<SymbolType> output) {
	std::variant<EpsilonType, SymbolType> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push), std::move(output));
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDTA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push, std::vector<SymbolType> output) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push), std::move(output));
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDTA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push, const std::vector<SymbolType>& output) {
	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> > key(from, input, pop);
	std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > value(to, push, output);

	return transitions[key].erase(value);
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDTA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const SymbolType& input, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push, const std::vector<SymbolType>& output) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push, output);
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDTA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push, const std::vector<SymbolType>& output) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeTransition(from, inputVariant, pop, to, push, output);
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > > >& NPDTA < SymbolType, EpsilonType, StateType >::getTransitions() const {
	return transitions;
}

template < class SymbolType, class EpsilonType, class StateType >
std::set < std::pair < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::tuple < StateType, std::vector < SymbolType >, std::vector < SymbolType > > > > NPDTA < SymbolType, EpsilonType, StateType >::getTransitionsFromState ( const StateType & from ) const {
	if( !getStates().count(from))
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist");

	std::set < std::pair < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::tuple < StateType, std::vector < SymbolType >, std::vector < SymbolType > > > > transitionsFromState;
	for ( auto transition: transitions ) {
		for ( auto iter: transition.second ) {
			if ( std::get<0>(transition.first) == from ) {
				transitionsFromState.insert(std::make_pair(transition.first, iter));
			}
		}
	}

	return transitionsFromState;
}

template < class SymbolType, class EpsilonType, class StateType >
int NPDTA < SymbolType, EpsilonType, StateType >::compare(const NPDTA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getOutputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getOutputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDTA < SymbolType, EpsilonType, StateType >::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "outputAlphabet = " << getOutputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

template < class SymbolType, class EpsilonType, class StateType >
NPDTA < SymbolType, EpsilonType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class EpsilonType, class StateType >
NPDTA < SymbolType, EpsilonType, StateType > NPDTA < SymbolType, EpsilonType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NPDTA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<SymbolType>(input);
	std::set<SymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	std::set<SymbolType> outputSymbols = AutomatonFromXMLParser::parseOutputAlphabet<SymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	SymbolType initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	NPDTA < SymbolType, EpsilonType, StateType > automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setOutputAlphabet(std::move(outputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<NPDTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NPDTA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDTA < SymbolType, EpsilonType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, NPDTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
	std::vector<SymbolType> pop = AutomatonFromXMLParser::parseTransitionPop<SymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	std::vector<SymbolType> push = AutomatonFromXMLParser::parseTransitionPush<SymbolType>(input);
	std::vector<SymbolType> output = AutomatonFromXMLParser::parseTransitionOutputSymbolMultiple<SymbolType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push), std::move(output));
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDTA < SymbolType, EpsilonType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NPDTA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeOutputAlphabet(out, this->getOutputAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(NPDTA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDTA < SymbolType, EpsilonType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionPop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, std::get<0>(target));
			AutomatonToXMLComposer::composeTransitionPush(out, std::get<1>(target));
			AutomatonToXMLComposer::composeTransitionOutputSymbolMultiple(out, std::get<2>(target));

			out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
alib::ObjectBase* NPDTA < SymbolType, EpsilonType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).template is<SymbolType>() && symbol == std::get<1>(transition.first).template get<SymbolType>())
				return true;

		return false;
	}

	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::OutputAlphabet > {
public:
	static bool used ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > > >& transition : automaton.getTransitions())
			for(const std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> >& target : transition.second)
				if (std::find(std::get<2>(target).begin(), std::get<2>(target).end(), symbol) != std::get<2>(target).end())
					return true;

		return false;
	}

	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getInitialSymbol() == symbol)
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > > >& transition : automaton.getTransitions()) {
			for (const SymbolType& popSymbol : std::get<2>(transition.first))
				if (symbol == popSymbol)
					return true;

			for (const std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> >& target : transition.second)
				if (std::find(std::get<1>(target).begin(), std::get<1>(target).end(), symbol) != std::get<1>(target).end())
					return true;
		}

		return false;
	}

	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> > > >& transition : automaton.getTransitions()) {
			if (state == std::get<0>(transition.first))
				return true;

			for(const std::tuple<StateType, std::vector<SymbolType>, std::vector<SymbolType> >& target : transition.second)
				if(std::get<0>(target) == state)
					return true;
		}

		return false;
	}

	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::NPDTA < SymbolType, EpsilonType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* NPDTA_H_ */
