/*
 * REAL_TIME_HEIGHT_DETERMINISTIC_NPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_
#define REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_

#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialStates;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
template < class SymbolType, class EpsilonType, class StateType >
class RealTimeHeightDeterministicNPDA : public AutomatonBase, public std::Components < RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < StateType, std::variant < EpsilonType, SymbolType > >, std::set < std::pair < StateType, SymbolType > > > callTransitions;
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, SymbolType >, std::set < StateType > > returnTransitions;
	std::map < std::pair < StateType, std::variant < EpsilonType, SymbolType > >, std::set < StateType > > localTransitions;

public:
	explicit RealTimeHeightDeterministicNPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreSymbol, std::set < StateType > initialStates, SymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates );
	explicit RealTimeHeightDeterministicNPDA ( SymbolType bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getInitialStates ( ) const {
		return this->template accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( StateType state ) {
		return this->template accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < StateType > states ) {
		this->template accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	void removeInitialState ( const StateType & state ) {
		this->template accessComponent < InitialStates > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getBottomOfTheStackSymbol ( ) const {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( SymbolType symbol ) {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( StateType current, std::variant < EpsilonType, SymbolType > input, StateType next, SymbolType push );
	bool addCallTransition ( StateType current, StateType next, SymbolType push );
	bool addCallTransition ( StateType current, SymbolType input, StateType next, SymbolType push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( StateType current, std::variant < EpsilonType, SymbolType > input, SymbolType pop, StateType next );
	bool addReturnTransition ( StateType current, SymbolType pop, StateType next );
	bool addReturnTransition ( StateType current, SymbolType input, SymbolType pop, StateType next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( StateType current, std::variant < EpsilonType, SymbolType > input, StateType next );
	bool addLocalTransition ( StateType current, StateType next );
	bool addLocalTransition ( StateType current, SymbolType input, StateType next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const StateType & current, const std::variant < EpsilonType, SymbolType > & input, const StateType & next, const SymbolType & push );
	bool removeCallTransition ( const StateType & current, const StateType & next, const SymbolType & push );
	bool removeCallTransition ( const StateType & current, const SymbolType & input, const StateType & next, const SymbolType & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const StateType & current, const std::variant < EpsilonType, SymbolType > & input, const SymbolType & pop, const StateType & next );
	bool removeReturnTransition ( const StateType & current, const SymbolType & pop, const StateType & next );
	bool removeReturnTransition ( const StateType & current, const SymbolType & input, const SymbolType & pop, const StateType & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const StateType & current, const std::variant < EpsilonType, SymbolType > & input, const StateType & next );
	bool removeLocalTransition ( const StateType & current, const StateType & next );
	bool removeLocalTransition ( const StateType & current, const SymbolType & input, const StateType & next );

	const std::map < std::pair < StateType, std::variant < EpsilonType, SymbolType > >, std::set < std::pair < StateType, SymbolType > > > & getCallTransitions ( ) const;

	const std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, SymbolType >, std::set < StateType > > & getReturnTransitions ( ) const;

	const std::map < std::pair < StateType, std::variant < EpsilonType, SymbolType > >, std::set < StateType > > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RealTimeHeightDeterministicNPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RealTimeHeightDeterministicNPDA";

		return xmlTagName;
	}

	static RealTimeHeightDeterministicNPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, RealTimeHeightDeterministicNPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class EpsilonType, class StateType >
RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::RealTimeHeightDeterministicNPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreAlphabet, std::set < StateType > initialStates, SymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates ) : std::Components < RealTimeHeightDeterministicNPDA, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( bottomOfTheStackSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( initialStates ), std::move ( finalStates ) ), std::tuple < > { } ) {
}

template < class SymbolType, class EpsilonType, class StateType >
RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::RealTimeHeightDeterministicNPDA(SymbolType bottomOfTheStackSymbol) : RealTimeHeightDeterministicNPDA ( std::set < StateType > { }, std::set < SymbolType > { }, std::set < SymbolType > { bottomOfTheStackSymbol }, std::set < StateType > {}, bottomOfTheStackSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::clone() const {
	return new RealTimeHeightDeterministicNPDA(*this);
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::plunder() && {
	return new RealTimeHeightDeterministicNPDA(std::move(*this));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addCallTransition(StateType from, std::variant<EpsilonType, SymbolType> input, StateType to, SymbolType push) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is<SymbolType>() && ! getInputAlphabet().count(input.template get<SymbolType>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(push)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( push ) + "\" doesn't exist.");
	}

	std::pair<StateType, std::variant<EpsilonType, SymbolType>> key(std::move(from), std::move(input));
	std::pair<StateType, SymbolType> value = std::make_pair(std::move(to), std::move(push));

	return callTransitions[std::move(key)].insert(std::move(value)).second;
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addCallTransition(StateType from, StateType to, SymbolType push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addCallTransition(StateType from, SymbolType input, StateType to, SymbolType push) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addReturnTransition(StateType from, std::variant<EpsilonType, SymbolType> input, SymbolType pop, StateType to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is<SymbolType>() && !getInputAlphabet().count(input.template get<SymbolType>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType> key(std::move(from), std::move(input), std::move(pop));

	return returnTransitions[std::move(key)].insert(std::move(to)).second;
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addReturnTransition(StateType from, SymbolType pop, StateType to) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addReturnTransition(StateType from, SymbolType input, SymbolType pop, StateType to) {
	std::variant<EpsilonType, SymbolType> inputVariant(std::move(input));
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addLocalTransition(StateType from, std::variant<EpsilonType, SymbolType> input, StateType to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is<SymbolType>() && ! getInputAlphabet().count(input.template get<SymbolType>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	std::pair<StateType, std::variant<EpsilonType, SymbolType>> key(std::move(from), std::move(input));

	return localTransitions[std::move(key)].insert(std::move(to)).second;
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addLocalTransition(StateType from, StateType to) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::addLocalTransition(StateType from, SymbolType input, StateType to) {
	std::variant<EpsilonType, SymbolType> inputVariant(std::move(input));
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeCallTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const StateType& to, const SymbolType& push) {
	std::pair<StateType, std::variant<EpsilonType, SymbolType>> key(from, input);
	std::pair<StateType, SymbolType> value = std::make_pair(to, push);

	return callTransitions[key].erase(value);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeCallTransition(const StateType& from, const StateType& to, const SymbolType& push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeCallTransition(from, inputVariant, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeCallTransition(const StateType& from, const SymbolType& input, const StateType& to, const SymbolType& push) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeCallTransition(from, inputVariant, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeReturnTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const SymbolType& pop, const StateType& to) {
	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType> key(from, input, pop);

	return returnTransitions[key].erase(to);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeReturnTransition(const StateType& from, const SymbolType& pop, const StateType& to) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeReturnTransition(from, inputVariant, pop, to);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeReturnTransition(const StateType& from, const SymbolType& input, const SymbolType& pop, const StateType& to) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeReturnTransition(from, inputVariant, pop, to);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeLocalTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const StateType& to) {
	std::pair<StateType, std::variant<EpsilonType, SymbolType>> key(from, input);

	return localTransitions[key].erase(to);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeLocalTransition(const StateType& from, const StateType& to) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeLocalTransition(from, inputVariant, to);
}

template < class SymbolType, class EpsilonType, class StateType >
bool RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::removeLocalTransition(const StateType& from, const SymbolType& input, const StateType& to) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeLocalTransition(from, inputVariant, to);
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<std::pair<StateType, SymbolType> > >& RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::getCallTransitions() const {
	return callTransitions;
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::set<StateType> >& RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::getReturnTransitions() const {
	return returnTransitions;
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<StateType> >& RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::getLocalTransitions() const {
	return localTransitions;
}

template < class SymbolType, class EpsilonType, class StateType >
int RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::compare(const RealTimeHeightDeterministicNPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialStates(), getFinalStates(), getPushdownStoreAlphabet(), getBottomOfTheStackSymbol(), callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialStates(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getBottomOfTheStackSymbol(), other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class EpsilonType, class StateType >
void RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::operator>>(std::ostream& out) const {
	out << "(RealTimeHeightDeterministicNPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialStates = " << getInitialStates()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "bottomOfTheStackSymbol = " << getBottomOfTheStackSymbol()
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

template < class SymbolType, class EpsilonType, class StateType >
RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class EpsilonType, class StateType >
RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RealTimeHeightDeterministicNPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<SymbolType>(input);
	std::set<SymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	std::set<StateType> initialStates = AutomatonFromXMLParser::parseInitialStates<StateType>(input);
	SymbolType bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	RealTimeHeightDeterministicNPDA automaton(std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setInitialStates(std::move(initialStates));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<RealTimeHeightDeterministicNPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RealTimeHeightDeterministicNPDA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class EpsilonType, class StateType >
void RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, RealTimeHeightDeterministicNPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
		SymbolType push = AutomatonFromXMLParser::parseTransitionSinglePush<SymbolType>(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
		SymbolType pop = AutomatonFromXMLParser::parseTransitionSinglePop<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

template < class SymbolType, class EpsilonType, class StateType >
void RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RealTimeHeightDeterministicNPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialStates(out, this->getInitialStates());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(RealTimeHeightDeterministicNPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
void RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionSinglePush(out, target.second);

			out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getReturnTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getLocalTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
alib::ObjectBase* RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<std::pair<StateType, SymbolType> > >& callTransition : automaton.getCallTransitions())
			if (callTransition.first.second.template is<SymbolType>() && symbol == callTransition.first.second.template get<SymbolType>())
				return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions())
			if (std::get<1>(returnTransition.first).template is<SymbolType>() && symbol == std::get<1>(returnTransition.first).template get<SymbolType>())
				return true;

		for (const std::pair<const std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<StateType> >& localTransition : automaton.getLocalTransitions())
			if (localTransition.first.second.template is<SymbolType>() && symbol == localTransition.first.second.template get<SymbolType>())
				return true;

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getBottomOfTheStackSymbol() == symbol)
			return true;

		for (const std::pair<const std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<std::pair<StateType, SymbolType> > >& callTransition : automaton.getCallTransitions())
			for(const std::pair<StateType, SymbolType>& to : callTransition.second)
				if (symbol == to.second)
					return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<2>(returnTransition.first))
				return true;

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::BottomOfTheStackSymbol > {
public:
	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialStates ( ).count ( state ) )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<std::pair<StateType, SymbolType> > >& callTransition : automaton.getCallTransitions()) {
			if (state == callTransition.first.first)
				return true;

			for(const std::pair<StateType, SymbolType>& target : callTransition.second)
				if(target.first == state)
					return true;
		}

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions()) {
			if (state == std::get<0>(returnTransition.first))
				return true;

			for(const StateType& target : returnTransition.second)
				if(target == state)
					return true;
		}

		for (const std::pair<const std::pair<StateType, std::variant<EpsilonType, SymbolType>>, std::set<StateType> >& localTransition : automaton.getLocalTransitions()) {
			if (state == localTransition.first.first)
				return true;

			for(const StateType& target : localTransition.second)
				if(target == state)
					return true;
		}

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::InitialStates > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_ */
