/*
 * SinglePopDPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef SINGLE_POP_DPDA_H_
#define SINGLE_POP_DPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>
#include <ostream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Automaton
 *
 * $|\delta (q, a, \gamma)| \leq 1$, $\forall q, a, \gamma, q \in Q, a \in (T \cup \{\varepsilon\}), \gamma \in G^*.$
 * if $\delta (q, a, r) \neq \emptyset$, $\delta (q, a, s) \neq \emptyset$ then $r \neq s$.
 * if $\delta(q, a, r) \neq \emptyset$, $\delta (q, \varepsilon, s) \neq \emptyset$, then $r \neq s$.
 */
template < class SymbolType, class EpsilonType, class StateType >
class SinglePopDPDA : public AutomatonBase, public std::Components < SinglePopDPDA < SymbolType, EpsilonType, StateType >, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, SymbolType >, std::pair < StateType, std::vector < SymbolType > > > transitions;

public:
	explicit SinglePopDPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreSymbol, StateType initialState, SymbolType initialPushdownSymbol, std::set < StateType > finalStates );
	explicit SinglePopDPDA ( StateType initialState, SymbolType initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the SinglePopDPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the SinglePopDPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, std::variant < EpsilonType, SymbolType > input, SymbolType pop, StateType to, std::vector < SymbolType > push );

	bool addTransition ( StateType from, SymbolType input, SymbolType pop, StateType to, std::vector < SymbolType > push );

	bool addTransition ( StateType from, SymbolType pop, StateType to, std::vector < SymbolType > push );

	/**
	 * Removes the transition from the SinglePopDPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the SinglePopDPDA
	 */
	bool removeTransition ( const StateType & from, const std::variant < EpsilonType, SymbolType > & input, const SymbolType & pop, const StateType & to, const std::vector < SymbolType > & push );

	bool removeTransition ( const StateType & from, const SymbolType & input, const SymbolType & pop, const StateType & to, const std::vector < SymbolType > & push );

	bool removeTransition ( const StateType & from, const SymbolType & pop, const StateType & to, const std::vector < SymbolType > & push );

	/**
	 * @return SinglePopDPDA transitions
	 */
	const std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, SymbolType >, std::pair < StateType, std::vector < SymbolType > > > & getTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SinglePopDPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "SinglePopDPDA";

		return xmlTagName;
	}

	static SinglePopDPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, SinglePopDPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class EpsilonType, class StateType >
SinglePopDPDA < SymbolType, EpsilonType, StateType >::SinglePopDPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreAlphabet, StateType initialState, SymbolType initialSymbol, std::set < StateType > finalStates ) : std::Components < SinglePopDPDA, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class SymbolType, class EpsilonType, class StateType >
SinglePopDPDA < SymbolType, EpsilonType, StateType >::SinglePopDPDA(StateType initialState, SymbolType initialPushdownSymbol) : SinglePopDPDA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, std::set < SymbolType > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* SinglePopDPDA < SymbolType, EpsilonType, StateType >::clone() const {
	return new SinglePopDPDA(*this);
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* SinglePopDPDA < SymbolType, EpsilonType, StateType >::plunder() && {
	return new SinglePopDPDA(std::move(*this));
}

template < class SymbolType, class EpsilonType, class StateType >
bool SinglePopDPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::variant<EpsilonType, SymbolType> input, SymbolType pop, StateType to, std::vector<SymbolType> push) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is<SymbolType>() && ! getInputAlphabet().count(input.template get<SymbolType>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	for(const SymbolType& pushSymbol : push) {
		if (!getPushdownStoreAlphabet().count(pushSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");
		}
	}

	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType> key(std::move(from), std::move(input), std::move(pop));
	std::pair<StateType, std::vector<SymbolType> > value = std::make_pair(std::move(to), std::move(push));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (input.template is<SymbolType>())
			throw AutomatonException("Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).template get<SymbolType>() ) + "\", \"" + std::to_string ( std::get<2>(key) ) + "\") -> ?? already exists.");
		else
			throw AutomatonException("Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).template get<EpsilonType>() ) + "\", \"" + std::to_string ( std::get<2>(key) ) + "\") -> ?? already exists.");
	}

	if(input.template is<EpsilonType>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::pair<StateType, std::vector<SymbolType> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key) && std::get<2>(transition.first) == std::get<2>(key)) return true;
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	} else {
		std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType> key2(std::get<0>(key), std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( ), std::get<2>(key));
		if(transitions.find(key2) != transitions.end())
			throw AutomatonException("Conflicting transition");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)) );
	return true;
}

template < class SymbolType, class EpsilonType, class StateType >
bool SinglePopDPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, SymbolType input, SymbolType pop, StateType to, std::vector<SymbolType> push) {
	std::variant<EpsilonType, SymbolType> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool SinglePopDPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, SymbolType pop, StateType to, std::vector<SymbolType> push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool SinglePopDPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const SymbolType& pop, const StateType& to, const std::vector<SymbolType>& push) {
	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType> key(from, input, pop);
	std::pair<StateType, std::vector<SymbolType> > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.template is<SymbolType>())
			throw AutomatonException("Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.template get<SymbolType>() ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");
		else
			throw AutomatonException("Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.template get<EpsilonType>() ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

template < class SymbolType, class EpsilonType, class StateType >
bool SinglePopDPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const SymbolType& input, const SymbolType& pop, const StateType& to, const std::vector<SymbolType>& push) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
bool SinglePopDPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const SymbolType& pop, const StateType& to, const std::vector<SymbolType>& push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::pair<StateType, std::vector<SymbolType> > >& SinglePopDPDA < SymbolType, EpsilonType, StateType >::getTransitions() const {
	return transitions;
}

template < class SymbolType, class EpsilonType, class StateType >
int SinglePopDPDA < SymbolType, EpsilonType, StateType >::compare(const SinglePopDPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class EpsilonType, class StateType >
void SinglePopDPDA < SymbolType, EpsilonType, StateType >::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

template < class SymbolType, class EpsilonType, class StateType >
SinglePopDPDA < SymbolType, EpsilonType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class EpsilonType, class StateType >
SinglePopDPDA < SymbolType, EpsilonType, StateType > SinglePopDPDA < SymbolType, EpsilonType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, SinglePopDPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<SymbolType>(input);
	std::set<SymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	SymbolType initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	SinglePopDPDA < SymbolType, EpsilonType, StateType > automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<SinglePopDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, SinglePopDPDA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class EpsilonType, class StateType >
void SinglePopDPDA < SymbolType, EpsilonType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, SinglePopDPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
	SymbolType pop = AutomatonFromXMLParser::parseTransitionSinglePop<SymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	std::vector<SymbolType> push = AutomatonFromXMLParser::parseTransitionPush<SymbolType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
void SinglePopDPDA < SymbolType, EpsilonType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(SinglePopDPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(SinglePopDPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
void SinglePopDPDA < SymbolType, EpsilonType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionPush(out, transition.second.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
alib::ObjectBase* SinglePopDPDA < SymbolType, EpsilonType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::pair<StateType, std::vector<SymbolType> > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).template is<SymbolType>() && symbol == std::get<1>(transition.first).template get<SymbolType>())
				return true;

		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::pair<StateType, std::vector<SymbolType> > >& transition : automaton.getTransitions())
			if (symbol == std::get<2>(transition.first) || std::find(transition.second.second.begin(), transition.second.second.end(), symbol) != transition.second.second.end())
				return true;

		if(automaton.getInitialSymbol() == symbol)
			return true;

		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, SymbolType>, std::pair<StateType, std::vector<SymbolType> > >& transition : automaton.getTransitions())
			if (state == std::get<0>(transition.first) || transition.second.first == state)
				return true;

		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::SinglePopDPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* SINGLE_POP_DPDA_H_ */
