/*
 * NFTA.h
 *
 *  Created on: Mar 21, 2015
 *      Author: Stepan Plachy
 */

#ifndef NFTA_H_
#define NFTA_H_

#include <map>
#include <vector>
#include <ostream>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class InputAlphabet;
class States;
class FinalStates;

/**
 * Represents Finite Tree Automaton.
 * Can store nondeterministic finite tree automaton without epsilon transitions.
 */
template < class SymbolType, class RankType, class StateType >
class NFTA : public AutomatonBase, public std::Components < NFTA < SymbolType, RankType, StateType >, std::ranked_symbol < SymbolType, RankType >, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < > > {
	std::map < std::pair < std::ranked_symbol < SymbolType, RankType >, std::vector < StateType > >, std::set < StateType > > transitions;

public:
	explicit NFTA ( );
	explicit NFTA ( std::set < StateType > states, std::set < std::ranked_symbol < SymbolType, RankType > > inputAlphabet, std::set < StateType > finalStates );
	explicit NFTA ( const DFTA < SymbolType, RankType, StateType > & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < std::ranked_symbol < SymbolType, RankType > > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( std::ranked_symbol < SymbolType, RankType > symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < std::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current symbol in node
	 * @param children states of children of current node
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( const std::ranked_symbol < SymbolType, RankType > & current, const std::vector < StateType > & children, const StateType & next );

	/**
	 * Removes transition from the automaton.
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const std::ranked_symbol < SymbolType, RankType > symbol, const std::vector < StateType > & states, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < std::ranked_symbol < SymbolType, RankType >, std::vector < StateType > >, std::set < StateType > > & getTransitions ( ) const { return transitions; }

	std::set < StateType > getTransitionRightSide ( const std::ranked_symbol < SymbolType, RankType > symbol, const std::vector < StateType > & states );

	/**
	 * Determines whether NFTA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c is epsilon free. Trivial for this class
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic ( ) const;

	unsigned transitionsSize ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NFTA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "NFTA";

		return xmlTagName;
	}

	static NFTA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, NFTA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

} /* namespace automaton */

#include "DFTA.h"

namespace automaton {

template < class SymbolType, class RankType, class StateType >
NFTA < SymbolType, RankType, StateType >::NFTA ( std::set < StateType > states, std::set < std::ranked_symbol < SymbolType, RankType > > inputAlphabet, std::set < StateType > finalStates ) : std::Components < NFTA, std::ranked_symbol < SymbolType, RankType >, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

template < class SymbolType, class RankType, class StateType >
NFTA < SymbolType, RankType, StateType >::NFTA() : NFTA ( std::set < StateType > { }, std::set < std::ranked_symbol < SymbolType, RankType > > { }, std::set < StateType > { } ) {
}

template < class SymbolType, class RankType, class StateType >
NFTA < SymbolType, RankType, StateType >::NFTA(const DFTA < SymbolType, RankType, StateType > & other) : NFTA ( other.getStates(), other.getInputAlphabet(), other.getFinalStates() ) {
	for(const auto& transition : other.getTransitions()) {
		transitions[transition.first].insert(transition.second);
	}
}

template < class SymbolType, class RankType, class StateType >
AutomatonBase* NFTA < SymbolType, RankType, StateType >::clone() const {
	return new NFTA(*this);
}

template < class SymbolType, class RankType, class StateType >
AutomatonBase* NFTA < SymbolType, RankType, StateType >::plunder() && {
	return new NFTA(std::move(*this));
}

template < class SymbolType, class RankType, class StateType >
bool NFTA < SymbolType, RankType, StateType >::addTransition(const std::ranked_symbol < SymbolType, RankType > & symbol, const std::vector<StateType> & prevStates, const StateType & next) {
	if (prevStates.size() != ( size_t ) symbol.getRank() )
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (! getInputAlphabet().count(symbol))
		throw AutomatonException("Input symbol \"" + std::to_string ( symbol ) + "\" doesn't exist.");

	if (! getStates().count(next))
		throw AutomatonException("State \"" + std::to_string ( next ) + "\" doesn't exist.");

	for (const StateType& it : prevStates) {
		if (! getStates().count(it))
			throw AutomatonException("State \"" + std::to_string ( it ) + "\" doesn't exist.");
	}

	std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> > key = std::make_pair(symbol, prevStates);
	return transitions[key].insert(next).second;
}

template < class SymbolType, class RankType, class StateType >
bool NFTA < SymbolType, RankType, StateType >::removeTransition(const std::ranked_symbol < SymbolType, RankType > symbol, const std::vector<StateType> & states, const StateType & next) {
	std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> > key = std::make_pair(symbol, states);
	return transitions[key].erase(next);
}

template < class SymbolType, class RankType, class StateType >
bool NFTA < SymbolType, RankType, StateType >::isDeterministic() const {
	for (const std::pair<const std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> >, std::set<StateType> >& t : transitions)
		if (t . second.size() != 1 || t . second.size() != 0)
			return false;

	return true;
}

template < class SymbolType, class RankType, class StateType >
unsigned NFTA < SymbolType, RankType, StateType >::transitionsSize() const {
	int res = 0;

	for(const auto& transition : transitions )
		res += transition.second.size();

	return res;
}

template < class SymbolType, class RankType, class StateType >
int NFTA < SymbolType, RankType, StateType >::compare(const NFTA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getFinalStates(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getFinalStates(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class RankType, class StateType >
void NFTA < SymbolType, RankType, StateType >::operator>>(std::ostream& out) const {
	out << "(NFTA "
		<< " states = " << getStates()
		<< " inputAlphabet = " << getInputAlphabet()
		<< " finalStates = " << getFinalStates()
		<< " transitions = " << transitions
		<< ")";
}

template < class SymbolType, class RankType, class StateType >
NFTA < SymbolType, RankType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class RankType, class StateType >
NFTA < SymbolType, RankType, StateType > NFTA < SymbolType, RankType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NFTA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<std::ranked_symbol < SymbolType, RankType >> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet<std::ranked_symbol < SymbolType, RankType >>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	NFTA < SymbolType, RankType, StateType > automaton;
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<NFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NFTA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class RankType, class StateType >
void NFTA < SymbolType, RankType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, NFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	std::ranked_symbol < SymbolType, RankType > inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<std::ranked_symbol < SymbolType, RankType >>(input);
	std::vector<StateType> from = AutomatonFromXMLParser::parseTransitionFromMultiple<StateType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

template < class SymbolType, class RankType, class StateType >
void NFTA < SymbolType, RankType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NFTA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(NFTA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class RankType, class StateType >
void NFTA < SymbolType, RankType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		for(const auto& targetState: transition.second) {
			out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, targetState);

			out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
		}
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

template < class SymbolType, class RankType, class StateType >
alib::ObjectBase* NFTA < SymbolType, RankType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class RankType, class StateType >
class ComponentConstraint< automaton::NFTA < SymbolType, RankType, StateType >, std::ranked_symbol < SymbolType, RankType >, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		for (const std::pair<const std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> >, std::set<StateType>>& t : automaton.getTransitions())
			if (t.first.first == symbol)
				return true;

		return false;
	}

	static bool available ( const automaton::NFTA < SymbolType, RankType, StateType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const automaton::NFTA < SymbolType, RankType, StateType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
	}
};

template < class SymbolType, class RankType, class StateType >
class ComponentConstraint< automaton::NFTA < SymbolType, RankType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> >, std::set<StateType>>& t : automaton.getTransitions())
			if(std::contains(t.first.second.begin(), t.first.second.end(), state ) || t . second.count ( state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::NFTA < SymbolType, RankType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::NFTA < SymbolType, RankType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class RankType, class StateType >
class ComponentConstraint< automaton::NFTA < SymbolType, RankType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::NFTA < SymbolType, RankType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NFTA < SymbolType, RankType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* NFTA_H_ */
