/*
 * NonContractingGrammar.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef NON_CONTRACTING_GRAMMAR_H_
#define NON_CONTRACTING_GRAMMAR_H_

#include <map>
#include <vector>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"

namespace grammar {

/**
 * Non contracting grammar. Type 1 in Chomsky hierarchy. Produces context sensitive language.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

template < class SymbolType >
class NonContractingGrammar : public GrammarBase, public std::Components < NonContractingGrammar < SymbolType >, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	std::map < std::vector < SymbolType >, std::set < std::vector < SymbolType > > > rules;
	bool generatesEpsilon;

public:
	explicit NonContractingGrammar ( SymbolType initialSymbol );

	explicit NonContractingGrammar ( std::set < SymbolType > nonTerminalSymbols, std::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( std::vector < SymbolType > leftHandSide, std::vector < SymbolType > rightHandSide );

	const std::map < std::vector < SymbolType >, std::set < std::vector < SymbolType > > > & getRules ( ) const;

	bool removeRule ( const std::vector < SymbolType > & leftHandSide, const std::vector < SymbolType > & rightHandSide );

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getNonterminalAlphabet ( ) const {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < SymbolType > & getTerminalAlphabet ( ) const {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void setGeneratesEpsilon ( bool genEps );
	bool getGeneratesEpsilon ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NonContractingGrammar & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "NonContractingGrammar";

		return xmlTagName;
	}

	static NonContractingGrammar parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, NonContractingGrammar & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType >
NonContractingGrammar < SymbolType >::NonContractingGrammar ( SymbolType initialSymbol ) : NonContractingGrammar ( std::set < SymbolType > { initialSymbol }, std::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
NonContractingGrammar < SymbolType >::NonContractingGrammar ( std::set < SymbolType > nonterminalAlphabet, std::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : std::Components < NonContractingGrammar, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

template < class SymbolType >
GrammarBase * NonContractingGrammar < SymbolType >::clone ( ) const {
	return new NonContractingGrammar ( * this );
}

template < class SymbolType >
GrammarBase * NonContractingGrammar < SymbolType >::plunder ( ) && {
	return new NonContractingGrammar ( std::move ( * this ) );
}

template < class SymbolType >
bool NonContractingGrammar < SymbolType >::addRule ( std::vector < SymbolType > leftHandSide, std::vector < SymbolType > rightHandSide ) {
	int lSize = leftHandSide.size ( );
	int rSize = rightHandSide.size ( );

	if ( lSize > rSize )
		throw GrammarException ( "Invalid size of right hand side of a rule" );

	if ( std::all_of ( leftHandSide.begin ( ), leftHandSide.end ( ), [this] ( const SymbolType symbol ) {
			return !getNonterminalAlphabet ( ).count ( symbol );
		} ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const SymbolType & symbol : leftHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + std::to_string ( symbol ) + "\" is not neither terminal nor nonterminal symbol" );

	for ( const SymbolType & symbol : rightHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + std::to_string ( symbol ) + "\" is not neither terminal nor nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

template < class SymbolType >
const std::map < std::vector < SymbolType >, std::set < std::vector < SymbolType > > > & NonContractingGrammar < SymbolType >::getRules ( ) const {
	return rules;
}

template < class SymbolType >
bool NonContractingGrammar < SymbolType >::removeRule ( const std::vector < SymbolType > & leftHandSide, const std::vector < SymbolType > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
void NonContractingGrammar < SymbolType >::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

template < class SymbolType >
bool NonContractingGrammar < SymbolType >::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

template < class SymbolType >
int NonContractingGrammar < SymbolType >::compare ( const NonContractingGrammar & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void NonContractingGrammar < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(NonContractingGrammar "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules
	    << "generatesEpsilon = " << generatesEpsilon << ")";
}

template < class SymbolType >
NonContractingGrammar < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
NonContractingGrammar < SymbolType > NonContractingGrammar < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, NonContractingGrammar::getXmlTagName() );

	std::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	std::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	NonContractingGrammar < SymbolType > grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, NonContractingGrammar::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void NonContractingGrammar < SymbolType >::parseRule ( std::deque < sax::Token >::iterator & input, NonContractingGrammar & grammar ) {
	std::vector < SymbolType > lhs = GrammarFromXMLParser::parseRuleLHS < SymbolType > ( input );
	std::vector < SymbolType > rhs = GrammarFromXMLParser::parseRuleRHS < SymbolType > ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void NonContractingGrammar < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( NonContractingGrammar::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, this->getGeneratesEpsilon ( ) );

	out.emplace_back ( NonContractingGrammar::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void NonContractingGrammar < SymbolType >::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* NonContractingGrammar < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace std {

template < class SymbolType >
class ComponentConstraint< grammar::NonContractingGrammar < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	static bool used ( const grammar::NonContractingGrammar < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const std::vector < SymbolType >, std::set < std::vector < SymbolType > > > & rule : grammar.getRules ( ) ) {
			if ( std::find ( rule.first.begin ( ), rule.first.end ( ), symbol ) != rule.first.end ( ) )
				return true;

			for ( const std::vector < SymbolType > & rhs : rule.second )
				if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
					return true;

		}

		return false;
	}

	static bool available ( const grammar::NonContractingGrammar < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::NonContractingGrammar < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + std::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

template < class SymbolType >
class ComponentConstraint< grammar::NonContractingGrammar < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	static bool used ( const grammar::NonContractingGrammar < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const std::vector < SymbolType >, std::set < std::vector < SymbolType > > > & rule : grammar.getRules ( ) ) {
			if ( std::find ( rule.first.begin ( ), rule.first.end ( ), symbol ) != rule.first.end ( ) )
				return true;

			for ( const std::vector < SymbolType > & rhs : rule.second )
				if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
					return true;

		}

		if ( grammar.template accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	static bool available ( const grammar::NonContractingGrammar < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::NonContractingGrammar < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + std::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

template < class SymbolType >
class ElementConstraint< grammar::NonContractingGrammar < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	static bool available ( const grammar::NonContractingGrammar < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const grammar::NonContractingGrammar < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* NON_CONTRACTING_GRAMMAR_H_ */
