/*
 * RightRG.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef RIGHT_RG_H_
#define RIGHT_RG_H_

#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"

namespace grammar {

/**
 * Right regular grammar in chomsky normal form. Type 2 in Chomsky hierarchy. Produces regular languages.
 * Definition is similar to all common definitions of regular grammars. Additionaly contains boolean signaling whether the grammar generates empty string or don't.
 * G = (N, T, P, S, E)
 * N = nonempty finite set of nonterminal symbols
 * T = finite set of terminal symbols - having this empty won't let grammar do much though
 * P = set of production rules of the form A -> aB or A -> a (where A, B \in N and a \in T)
 * S = initial nonterminal symbol
 * E = boolean signaling wheter grammar generates empty string or don't
 *
 * This definition has simplier handling of empty string generation and it is compatible with common definitions where the transformation from this definition to the common definition and backwards is trivial
 */

class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

template < class SymbolType >
class RightRG : public GrammarBase, public std::Components < RightRG < SymbolType >, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	/**
	 * Transition function as mapping from nonterminal symbol on the left hand side to set of either terminal symbols or doublets of terminal symbol and nonterminal symbol
	 */
	std::map < SymbolType, std::set < std::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > rules;

	/**
	 * Boolean signaling whether grammar generates empty string or don't.
	 */
	bool generatesEpsilon;

public:
	explicit RightRG ( SymbolType initialSymbol );

	explicit RightRG ( std::set < SymbolType > nonTerminalSymbols, std::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( SymbolType leftHandSide, std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rightHandSide );
	bool addRule ( SymbolType leftHandSide, SymbolType rightHandSide );
	bool addRule ( SymbolType leftHandSide, std::pair < SymbolType, SymbolType > rightHandSide );

	const std::map < SymbolType, std::set < std::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & getRules ( ) const;

	bool removeRule ( const SymbolType & leftHandSide, const std::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rightHandSide );
	bool removeRule ( const SymbolType & leftHandSide, const SymbolType & rightHandSide );
	bool removeRule ( const SymbolType & leftHandSide, const std::pair < SymbolType, SymbolType > & rightHandSide );

	bool addRawRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide );

	std::map < SymbolType, std::set < std::vector < SymbolType > > > getRawRules ( ) const;

	bool removeRawRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide );

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getNonterminalAlphabet ( ) const {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < SymbolType > & getTerminalAlphabet ( ) const {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void setGeneratesEpsilon ( bool genEps );
	bool getGeneratesEpsilon ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RightRG & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RightRG";

		return xmlTagName;
	}

	static RightRG parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, RightRG & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType >
RightRG < SymbolType >::RightRG ( SymbolType initialSymbol ) : RightRG ( std::set < SymbolType > { initialSymbol }, std::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
RightRG < SymbolType >::RightRG ( std::set < SymbolType > nonterminalAlphabet, std::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : std::Components < RightRG, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

template < class SymbolType >
GrammarBase * RightRG < SymbolType >::clone ( ) const {
	return new RightRG ( * this );
}

template < class SymbolType >
GrammarBase * RightRG < SymbolType >::plunder ( ) && {
	return new RightRG ( std::move ( * this ) );
}

template < class SymbolType >
bool RightRG < SymbolType >::addRule ( SymbolType leftHandSide, std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( rightHandSide.template is < SymbolType > ( ) ) {
		const SymbolType & rhs = rightHandSide.template get < SymbolType > ( );

		if ( !getTerminalAlphabet ( ).count ( rhs ) )
			throw GrammarException ( "Rule must rewrite to terminal symbol" );

		return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
	} else {
		const std::pair < SymbolType, SymbolType > & rhs = rightHandSide.template get < std::pair < SymbolType, SymbolType > > ( );

		if ( !getTerminalAlphabet ( ).count ( rhs.first ) || !getNonterminalAlphabet ( ).count ( rhs.second ) )
			throw GrammarException ( "Rule must rewrite to terminal symbol followed by nonterminal symbol" );

		return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
	}
}

template < class SymbolType >
bool RightRG < SymbolType >::addRule ( SymbolType leftHandSide, SymbolType rightHandSide ) {
	std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
bool RightRG < SymbolType >::addRule ( SymbolType leftHandSide, std::pair < SymbolType, SymbolType > rightHandSide ) {
	std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
const std::map < SymbolType, std::set < std::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & RightRG < SymbolType >::getRules ( ) const {
	return rules;
}

template < class SymbolType >
bool RightRG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const std::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
bool RightRG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const SymbolType & rightHandSide ) {
	std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool RightRG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const std::pair < SymbolType, SymbolType > & rightHandSide ) {
	std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool RightRG < SymbolType >::addRawRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( true );
		return res;
	} else if ( rightHandSide.size ( ) == 1 ) {
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide[0] ) );
	} else if ( rightHandSide.size ( ) == 2 ) {
		return addRule ( std::move ( leftHandSide ), std::make_pair ( std::move ( rightHandSide[0] ), std::move ( rightHandSide[1] ) ) );
	} else {
		throw GrammarException ( "Invalid right hand side" );
	}
}

template < class SymbolType >
std::map < SymbolType, std::set < std::vector < SymbolType > > > RightRG < SymbolType >::getRawRules ( ) const {
	std::map < SymbolType, std::set < std::vector < SymbolType > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			if ( rhs.template is < SymbolType > ( ) ) {
				std::vector < SymbolType > tmp { rhs.template get < SymbolType > ( ) };
				res[rule.first].insert ( std::move ( tmp ) );
			} else {
				const auto & rhsPair = rhs.template get < std::pair < SymbolType, SymbolType > > ( );
				std::vector < SymbolType > tmp { rhsPair.first, rhsPair.second };
				res[rule.first].insert ( std::move ( tmp ) );
			}
		}

	if ( getGeneratesEpsilon ( ) )
		res[getInitialSymbol ( )].insert ( std::vector < SymbolType > { } );

	return res;
}

template < class SymbolType >
bool RightRG < SymbolType >::removeRawRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( false );
		return res;
	} else if ( rightHandSide.size ( ) == 1 ) {
		return removeRule ( leftHandSide, rightHandSide[0] );
	} else if ( rightHandSide.size ( ) == 2 ) {
		return removeRule ( leftHandSide, std::make_pair ( rightHandSide[0], rightHandSide[1] ) );
	} else {
		throw GrammarException ( "Invalid right hand side" );
	}
}

template < class SymbolType >
void RightRG < SymbolType >::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

template < class SymbolType >
bool RightRG < SymbolType >::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

template < class SymbolType >
int RightRG < SymbolType >::compare ( const RightRG & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void RightRG < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(RightRG "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules
	    << "generatesEpsilon = " << generatesEpsilon << ")";
}

template < class SymbolType >
RightRG < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
RightRG < SymbolType > RightRG < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RightRG::getXmlTagName() );

	std::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	std::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	RightRG < SymbolType > grammar ( std::move ( nonterminalAlphabet ), std::move ( terminalAlphabet ), std::move ( initialSymbol ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );

	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RightRG::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void RightRG < SymbolType >::parseRule ( std::deque < sax::Token >::iterator & input, RightRG & grammar ) {
	SymbolType lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS < SymbolType > ( input );
	std::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs = GrammarFromXMLParser::parseRuleOneOrTwoSymbolsRHS < SymbolType > ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void RightRG < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RightRG::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, getInitialSymbol ( ) );
	composeRules ( out );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, getGeneratesEpsilon ( ) );

	out.emplace_back ( RightRG::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void RightRG < SymbolType >::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleOneOrTwoSymbolsRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* RightRG < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace std {

template < class SymbolType >
class ComponentConstraint< grammar::RightRG < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	static bool used ( const grammar::RightRG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::set < std::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & rule : grammar.getRules ( ) )
			for ( const std::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rhs : rule.second )
				if ( ( rhs.template is < SymbolType > ( ) && ( rhs.template get < SymbolType > ( ) == symbol ) ) || ( rhs.template get < std::pair < SymbolType, SymbolType > > ( ).first == symbol ) )
					return true;

		return false;
	}

	static bool available ( const grammar::RightRG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::RightRG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + std::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

template < class SymbolType >
class ComponentConstraint< grammar::RightRG < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	static bool used ( const grammar::RightRG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::set < std::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & rule : grammar.getRules ( ) ) {
			if ( rule.first == symbol )
				return true;

			for ( const std::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rhs : rule.second )
				if ( rhs.template get < std::pair < SymbolType, SymbolType > > ( ).second == symbol )
					return true;

		}

		if ( grammar.template accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	static bool available ( const grammar::RightRG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::RightRG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + std::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

template < class SymbolType >
class ElementConstraint< grammar::RightRG < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	static bool available ( const grammar::RightRG < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const grammar::RightRG < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* RIGHT_RG_H_ */
