/*
 * SuffixTrie.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef SUFFIX_TRIE_H_
#define SUFFIX_TRIE_H_

#include <string>
#include <set>
#include <trie>
#include <variant>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/Object.h>
#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>
#include <primitive/Bool.h>

#include <container/ObjectsSet.h>
#include <container/ObjectsTrie.h>
#include <container/ObjectsVariant.h>
#include <object/Void.h>

namespace indexes {

class EdgeAlphabet;
class NodeAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class EdgeType = alib::Object, class NodeType = alib::Object >
class SuffixTrie : public alib::ObjectBase, public std::Components < SuffixTrie < EdgeType, NodeType >, EdgeType, std::tuple < EdgeAlphabet >, std::tuple < >, NodeType, std::tuple < NodeAlphabet >, std::tuple < > > {
protected:
	std::trie < EdgeType, std::variant < void, NodeType > > m_trie;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit SuffixTrie ( std::set < EdgeType > edgeAlphabet, std::set < NodeType > nodeAlphabet );
	explicit SuffixTrie ( std::set < EdgeType > edgeAlphabet, std::set < NodeType > nodeAlphabet, std::trie < EdgeType, std::variant < void, NodeType > > trie );
	explicit SuffixTrie ( std::trie < EdgeType, std::variant < void, NodeType > > trie );

	void checkTrie ( const std::trie < EdgeType, std::variant < void, NodeType > > & trie );

	/**
	 * @return Root node of the trie
	 */
	const std::trie < EdgeType, std::variant < void, NodeType > > & getRoot ( ) const;

	const std::set < EdgeType > & getEdgeAlphabet ( ) const {
		return this->template accessComponent < EdgeAlphabet > ( ).get ( );
	}

	const std::set < NodeType > & getNodeAlphabet ( ) const {
		return this->template accessComponent < NodeAlphabet > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( std::trie < EdgeType, std::variant < void, NodeType > > tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromEdgeAlphabet ( const EdgeType & symbol );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromNodeAlphabet ( const NodeType & symbol );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SuffixTrie & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "SuffixTrie";

		return xmlTagName;
	}

	static SuffixTrie parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

} /* namespace indexes */

namespace indexes {

template < class EdgeType, class NodeType >
SuffixTrie < EdgeType, NodeType >::SuffixTrie ( std::set < EdgeType > edgeAlphabet, std::set < NodeType > nodeAlphabet ) : SuffixTrie ( std::move ( edgeAlphabet ), std::move ( nodeAlphabet ), std::trie < EdgeType, std::variant < void, NodeType > > ( std::variant < void, NodeType >::template from < void > ( ) ) ) {
}

template < class EdgeType, class NodeType >
SuffixTrie < EdgeType, NodeType >::SuffixTrie ( std::set < EdgeType > edgeAlphabet, std::set < NodeType > nodeAlphabet, std::trie < EdgeType, std::variant < void, NodeType > > trie ) : std::Components < SuffixTrie, EdgeType, std::tuple < EdgeAlphabet >, std::tuple < >, NodeType, std::tuple < NodeAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( edgeAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( nodeAlphabet ) ), std::tuple < > ( ) ), m_trie ( std::move ( trie ) ) {
	checkTrie ( this->m_trie );
}

template < class EdgeType, class NodeType >
SuffixTrie < EdgeType, NodeType >::SuffixTrie ( std::trie < EdgeType, std::variant < void, NodeType > > trie ) : SuffixTrie ( computeMinimalEdgeAlphabet ( trie ), computeMinimalNodeAlphabet ( trie ), trie ) {
}

template < class EdgeType, class NodeType >
alib::ObjectBase * SuffixTrie < EdgeType, NodeType >::clone ( ) const {
	return new SuffixTrie ( * this );
}

template < class EdgeType, class NodeType >
alib::ObjectBase * SuffixTrie < EdgeType, NodeType >::plunder ( ) && {
	return new SuffixTrie ( std::move ( * this ) );
}

template < class EdgeType, class NodeType >
void SuffixTrie < EdgeType, NodeType >::checkTrie ( const std::trie < EdgeType, std::variant < void, NodeType > > & trie ) {
	if ( trie.getData ( ).template is < NodeType > ( ) && ! getNodeAlphabet ( ).count ( trie.getData ( ).template get < NodeType > ( ) ) )
		throw exception::CommonException ( "Node symbols not in the node alphabet." );

	for ( const std::pair < const EdgeType, std::trie < EdgeType, std::variant < void, NodeType > > > & child : trie.getChildren ( ) ) {
		if ( ! getEdgeAlphabet ( ).count ( child.first ) )
			throw exception::CommonException ( "Node symbols not in the node alphabet." );
		checkTrie ( child.second );
	}
}

template < class EdgeType, class NodeType >
const std::trie < EdgeType, std::variant < void, NodeType > > & SuffixTrie < EdgeType, NodeType >::getRoot ( ) const {
	return m_trie;
}

template < class EdgeType, class NodeType >
void SuffixTrie < EdgeType, NodeType >::setTree ( std::trie < EdgeType, std::variant < void, NodeType > > trie ) {
	checkTrie ( trie );
	this->m_trie = std::move ( trie ).plunder ( );
}

template < class EdgeType, class NodeType >
void SuffixTrie < EdgeType, NodeType >::operator >>( std::ostream & out ) const {
	out << "(SuffixTrie " << this->m_trie << ")";
}

template < class EdgeType, class NodeType >
int SuffixTrie < EdgeType, NodeType >::compare ( const SuffixTrie & other ) const {
	auto first = std::tie ( getRoot ( ), getEdgeAlphabet ( ), getNodeAlphabet ( ) );
	auto second = std::tie ( other.getRoot ( ), other.getEdgeAlphabet ( ), getNodeAlphabet ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class EdgeType, class NodeType >
SuffixTrie < EdgeType, NodeType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class EdgeType, class NodeType >
SuffixTrie < EdgeType, NodeType > SuffixTrie < EdgeType, NodeType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixTrie::getXmlTagName() );
	std::set < EdgeType > edgeAlphabet = alib::xmlApi < std::set < EdgeType > >::parse ( input );
	std::set < NodeType > nodeAlphabet = alib::xmlApi < std::set < NodeType > >::parse ( input );
	std::trie < EdgeType, std::variant < void, NodeType > > root = alib::xmlApi < std::trie < EdgeType, std::variant < void, NodeType > > >::parse ( input );
	SuffixTrie < EdgeType, NodeType > trie ( std::move ( edgeAlphabet ), std::move ( nodeAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixTrie::getXmlTagName() );
	return trie;
}

template < class EdgeType, class NodeType >
void SuffixTrie < EdgeType, NodeType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixTrie::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < std::set < EdgeType > >::compose ( out, getEdgeAlphabet ( ) );
	alib::xmlApi < std::set < NodeType > >::compose ( out, getNodeAlphabet ( ) );
	alib::xmlApi < std::trie < EdgeType, std::variant < void, NodeType > > >::compose ( out, getRoot ( ) );
	out.emplace_back ( SuffixTrie::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class EdgeType, class NodeType >
alib::ObjectBase* SuffixTrie < EdgeType, NodeType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace indexes */

namespace std {

template < class EdgeType, class NodeType >
class ComponentConstraint < indexes::SuffixTrie < EdgeType, NodeType >, EdgeType, indexes::EdgeAlphabet > {

	static bool used ( const std::trie < EdgeType, std::variant < void, NodeType > > & trie, const EdgeType & symbol ) {
		for ( const std::pair < const EdgeType, std::trie < EdgeType, std::variant < void, NodeType > > > & child : trie.getChildren ( ) ) {
			if ( symbol == child.first || checkTrie ( trie, child.second ) )
				return true;
		}
		return false;
	}

public:
	static bool used ( const indexes::SuffixTrie < EdgeType, NodeType > & index, const EdgeType & symbol ) {
		return used ( index.getRoot ( ), symbol );
	}

	static bool available ( const indexes::SuffixTrie < EdgeType, NodeType > &, const EdgeType & ) {
		return true;
	}

	static void valid ( const indexes::SuffixTrie < EdgeType, NodeType > &, const EdgeType & ) {
	}
};

template < class EdgeType, class NodeType >
class ComponentConstraint < indexes::SuffixTrie < EdgeType, NodeType >, NodeType, indexes::NodeAlphabet > {

	static bool used ( const std::trie < EdgeType, std::variant < void, NodeType > > & trie, const NodeType & symbol ) {
		if ( trie.getData ( ).template is < NodeType > ( ) && symbol == trie.getData ( ).template get < NodeType > ( ) )
			return true;

		for ( const std::pair < const EdgeType, std::trie < EdgeType, std::variant < void, NodeType > > > & child : trie.getChildren ( ) ) {
			if ( used ( trie, child.second ) )
				return true;
		}
		return false;
	}

public:
	static bool used ( const indexes::SuffixTrie < EdgeType, NodeType > & index, const NodeType & symbol ) {
		return used ( index.getRoot ( ), symbol );
	}

	static bool available ( const indexes::SuffixTrie < EdgeType, NodeType > &, const NodeType & ) {
		return true;
	}

	static void valid ( const indexes::SuffixTrie < EdgeType, NodeType > &, const NodeType & ) {
	}
};

} /* namespace std */

#endif /* SUFFIX_TRIE_H_ */
