/*
 * FormalRegExpIteration.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef FORMAL_REG_EXP_ITERATION_H_
#define FORMAL_REG_EXP_ITERATION_H_

#include <sstream>

#include <exception/CommonException.h>

#include "FormalRegExpElement.h"

namespace regexp {

/**
 * Represents iteration operator in the regular expression. Contains one FormalRegExpElement
 * as operand.
 */
template < class SymbolType >
class FormalRegExpIteration : public FormalRegExpElement < SymbolType >, public std::UnaryNode < std::smart_ptr < FormalRegExpElement < SymbolType > >, std::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpIteration < SymbolType > > {
public:
	void accept ( typename FormalRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit FormalRegExpIteration ( FormalRegExpElement < SymbolType > && );
	explicit FormalRegExpIteration ( const FormalRegExpElement < SymbolType > & );

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	FormalRegExpIteration < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	FormalRegExpIteration < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	std::smart_ptr < UnboundedRegExpElement < SymbolType > > asUnbounded ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc FormalRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc FormalRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const override;

	/**
	 * @return element
	 */
	const FormalRegExpElement < SymbolType > & getElement ( ) const;

	/**
	 * @return element
	 */
	FormalRegExpElement < SymbolType > & getElement ( );

	/**
	 * @param element to iterate
	 */
	void setElement ( FormalRegExpElement < SymbolType > && element );
	void setElement ( const FormalRegExpElement < SymbolType > & element );

	int compare ( const FormalRegExpElement < SymbolType > & other ) const override {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	int compare ( const FormalRegExpIteration & ) const;

	/**
	 * @copydoc FormalRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	explicit operator std::string ( ) const override;
};

} /* namespace regexp */

#include "FormalRegExpEmpty.h"
#include "../unbounded/UnboundedRegExpIteration.h"

namespace regexp {

template < class SymbolType >
FormalRegExpIteration < SymbolType >::FormalRegExpIteration ( FormalRegExpElement < SymbolType > && element ) : std::UnaryNode < std::smart_ptr < FormalRegExpElement < SymbolType > >, std::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpIteration < SymbolType > > ( std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) ) {
}

template < class SymbolType >
FormalRegExpIteration < SymbolType >::FormalRegExpIteration ( const FormalRegExpElement < SymbolType > & element ) : FormalRegExpIteration ( std::move_copy ( element ) ) {
}

template < class SymbolType >
const FormalRegExpElement < SymbolType > & FormalRegExpIteration < SymbolType >::getElement ( ) const {
	return * this->getChild ( );
}

template < class SymbolType >
FormalRegExpElement < SymbolType > & FormalRegExpIteration < SymbolType >::getElement ( ) {
	return * this->getChild ( );
}

template < class SymbolType >
void FormalRegExpIteration < SymbolType >::setElement ( FormalRegExpElement < SymbolType > && elementParam ) {
	this->setChild ( std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( elementParam ).plunder ( ) ) );
}

template < class SymbolType >
void FormalRegExpIteration < SymbolType >::setElement ( const FormalRegExpElement < SymbolType > & elementParam ) {
	setElement ( std::move_copy ( elementParam ) );
}

template < class SymbolType >
FormalRegExpIteration < SymbolType > * FormalRegExpIteration < SymbolType >::clone ( ) const {
	return new FormalRegExpIteration ( * this );
}

template < class SymbolType >
FormalRegExpIteration < SymbolType > * FormalRegExpIteration < SymbolType >::plunder ( ) && {
	return new FormalRegExpIteration ( std::move ( * this ) );
}

template < class SymbolType >
std::smart_ptr < UnboundedRegExpElement < SymbolType > > FormalRegExpIteration < SymbolType >::asUnbounded ( ) const {
	return std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( new UnboundedRegExpIteration < SymbolType > ( * getElement ( ).asUnbounded ( ) ) );
}

template < class SymbolType >
int FormalRegExpIteration < SymbolType >::compare ( const FormalRegExpIteration < SymbolType > & other ) const {
	return getElement ( ).compare ( other.getElement ( ) );
}

template < class SymbolType >
void FormalRegExpIteration < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(RegExpFormalRegExpIteration " << getElement ( ) << ")";
}

template < class SymbolType >
bool FormalRegExpIteration < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	return getElement ( ).testSymbol ( symbol );
}

template < class SymbolType >
void FormalRegExpIteration < SymbolType >::computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const {
	getElement ( ).computeMinimalAlphabet ( alphabet );
}

template < class SymbolType >
bool FormalRegExpIteration < SymbolType >::checkAlphabet ( const std::set < SymbolType > & alphabet ) const {
	return getElement ( ).checkAlphabet ( alphabet );
}

template < class SymbolType >
FormalRegExpIteration < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_ITERATION_H_ */
