/*
 * UnboundedRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_H_
#define UNBOUNDED_REG_EXP_H_

#include <string>
#include <set>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <exception/CommonException.h>

#include "../RegExpBase.h"
#include "../RegExpFeatures.h"
#include "../common/RegExpFromXMLParser.h"
#include "../common/RegExpToXMLComposer.h"

#include "UnboundedRegExpStructure.h"

namespace regexp {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType >
class UnboundedRegExp : public RegExpBase, public std::Components < UnboundedRegExp < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	UnboundedRegExpStructure < SymbolType > m_regExp;

public:
	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	virtual RegExpBase * clone ( ) const;

	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	virtual RegExpBase * plunder ( ) &&;

	explicit UnboundedRegExp ( );

	explicit UnboundedRegExp ( std::set < SymbolType > alphabet, UnboundedRegExpStructure < SymbolType > regExp );
	explicit UnboundedRegExp ( UnboundedRegExpStructure < SymbolType > regExp );
	explicit UnboundedRegExp ( const FormalRegExp < SymbolType > & other );

	/**
	 * @return Root node of the regular expression tree
	 */
	const UnboundedRegExpStructure < SymbolType > & getRegExp ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	UnboundedRegExpStructure < SymbolType > & getRegExp ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp ( UnboundedRegExpStructure < SymbolType > regExp );

	virtual const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnboundedRegExp & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnboundedRegExp";

		return xmlTagName;
	}

	static UnboundedRegExp parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

} /* namespace regexp */

#include "../formal/FormalRegExp.h"

namespace regexp {

template < class SymbolType >
UnboundedRegExp < SymbolType >::UnboundedRegExp ( std::set < SymbolType > alphabet, UnboundedRegExpStructure < SymbolType > regExp ) : std::Components < UnboundedRegExp, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), m_regExp ( std::move ( regExp ) ) {
	if ( !this->m_regExp.getStructure ( ).checkAlphabet ( getAlphabet ( ) ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType >
UnboundedRegExp < SymbolType >::UnboundedRegExp ( ) : UnboundedRegExp ( std::set < SymbolType > ( ), UnboundedRegExpStructure < SymbolType > ( ) ) {
}

template < class SymbolType >
UnboundedRegExp < SymbolType >::UnboundedRegExp ( UnboundedRegExpStructure < SymbolType > regExp ) : UnboundedRegExp ( regExp.getStructure ( ).computeMinimalAlphabet ( ), regExp ) {
}

template < class SymbolType >
UnboundedRegExp < SymbolType >::UnboundedRegExp ( const FormalRegExp < SymbolType > & other ) : UnboundedRegExp ( other.getAlphabet ( ), UnboundedRegExpStructure < SymbolType > ( other.getRegExp ( ) ) ) {
}

template < class SymbolType >
RegExpBase * UnboundedRegExp < SymbolType >::clone ( ) const {
	return new UnboundedRegExp ( * this );
}

template < class SymbolType >
RegExpBase * UnboundedRegExp < SymbolType >::plunder ( ) && {
	return new UnboundedRegExp ( std::move ( * this ) );
}

template < class SymbolType >
const UnboundedRegExpStructure < SymbolType > & UnboundedRegExp < SymbolType >::getRegExp ( ) const {
	return m_regExp;
}

template < class SymbolType >
UnboundedRegExpStructure < SymbolType > & UnboundedRegExp < SymbolType >::getRegExp ( ) {
	return m_regExp;
}

template < class SymbolType >
void UnboundedRegExp < SymbolType >::setRegExp ( UnboundedRegExpStructure < SymbolType > param ) {
	if ( !param.getStructure ( ).checkAlphabet ( getAlphabet ( ) ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );

	this->m_regExp = std::move ( param );
}

template < class SymbolType >
void UnboundedRegExp < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(UnboundedRegExp " << this->m_regExp.getStructure ( ) << ")";
}

template < class SymbolType >
int UnboundedRegExp < SymbolType >::compare ( const UnboundedRegExp & other ) const {
	auto first = std::tie ( m_regExp.getStructure ( ), getAlphabet ( ) );
	auto second = std::tie ( other.m_regExp.getStructure ( ), other.getAlphabet ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
UnboundedRegExp < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
UnboundedRegExp < SymbolType > UnboundedRegExp < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnboundedRegExp::getXmlTagName() );

	std::set < SymbolType > alphabet = RegExpFromXMLParser::parseAlphabet < SymbolType > ( input );

	UnboundedRegExpStructure < SymbolType > structure ( alib::xmlApi < regexp::UnboundedRegExpStructure < SymbolType > >::parse ( input ) );
	UnboundedRegExp < SymbolType > regexp ( std::move ( alphabet ), std::move ( structure ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnboundedRegExp::getXmlTagName() );

	return regexp;
}

template < class SymbolType >
void UnboundedRegExp < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnboundedRegExp::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	RegExpToXMLComposer::composeAlphabet ( out, this->getAlphabet ( ) );
	alib::xmlApi < UnboundedRegExpStructure < SymbolType > >::compose ( out, this->getRegExp ( ) );
	out.emplace_back ( UnboundedRegExp::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* UnboundedRegExp < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace regexp */

namespace std {

template < class SymbolType >
class ComponentConstraint< regexp::UnboundedRegExp < SymbolType >, SymbolType, regexp::GeneralAlphabet > {
public:
	static bool used ( const regexp::UnboundedRegExp < SymbolType > & regexp, const SymbolType & symbol ) {
		return regexp.getRegExp ( ).getStructure ( ).testSymbol ( symbol );
	}

	static bool available ( const regexp::UnboundedRegExp < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const regexp::UnboundedRegExp < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* UNBOUNDED_REG_EXP_H_ */
