/*
 * UnboundedRegExpStructure.h
 *
 *  Created on: Oct 20, 2016
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_STRUCTURE_H_
#define UNBOUNDED_REG_EXP_STRUCTURE_H_

#include <exception/CommonException.h>

#include "../RegExpFeatures.h"
#include "../common/RegExpFromXMLParser.h"
#include "../common/RegExpToXMLComposer.h"

#include "UnboundedRegExpEmpty.h"

namespace regexp {

/**
 * Represents unbounded regular expression structure as parsed from the XML. Regular expression is stored
 * as a tree of UnboundedRegExpElement.
 */
template < class SymbolType >
class UnboundedRegExpStructure {
protected:
	std::smart_ptr < UnboundedRegExpElement < SymbolType > > m_structure;

public:
	explicit UnboundedRegExpStructure ( );

	explicit UnboundedRegExpStructure ( UnboundedRegExpElement < SymbolType > && structure );
	explicit UnboundedRegExpStructure ( const UnboundedRegExpElement < SymbolType > & structure );
	explicit UnboundedRegExpStructure ( const FormalRegExpStructure < SymbolType > & other );

	/**
	 * @return Root node of the unbounded regular expression tree
	 */
	const UnboundedRegExpElement < SymbolType > & getStructure ( ) const;

	/**
	 * @return Root node of the unbounded regular expression tree
	 */
	UnboundedRegExpElement < SymbolType > & getStructure ( );

	/**
	 * Sets the root node of the unbounded regular expression tree
	 * @param structure root node to set
	 */
	void setStructure ( UnboundedRegExpElement < SymbolType > && structure );
	void setStructure ( const UnboundedRegExpElement < SymbolType > & structure );

	friend std::ostream & operator << ( std::ostream & out, const UnboundedRegExpStructure < SymbolType > & structure ) {
		out << structure.getStructure ( );
		return out;
	}

	friend bool operator < ( const UnboundedRegExpStructure & first, const UnboundedRegExpStructure & second ) {
		return first.getStructure().compare(second.getStructure()) < 0;
	}

	friend bool operator == ( const UnboundedRegExpStructure & first, const UnboundedRegExpStructure & second ) {
		return first.getStructure().compare(second.getStructure()) == 0;
	}
};

} /* namespace regexp */

namespace alib {

template < class SymbolType >
struct xmlApi < regexp::UnboundedRegExpStructure < SymbolType > > {
	static regexp::UnboundedRegExpStructure < SymbolType > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::UnboundedRegExpStructure < SymbolType > & data );
};

} /* namespace alib */

namespace std {

template < class SymbolType >
struct compare < regexp::UnboundedRegExpStructure < SymbolType > > {
	int operator()(const regexp::UnboundedRegExpStructure < SymbolType > & first, const regexp::UnboundedRegExpStructure < SymbolType > & second) const {
		return first.getStructure().compare(second.getStructure());
	}
};

} /* namespace std */

#include "../formal/FormalRegExpStructure.h"

namespace regexp {

template < class SymbolType >
UnboundedRegExpStructure < SymbolType >::UnboundedRegExpStructure ( UnboundedRegExpElement < SymbolType > && structure ) : m_structure ( NULL ) {
	setStructure ( std::move ( structure ) );
}

template < class SymbolType >
UnboundedRegExpStructure < SymbolType >::UnboundedRegExpStructure ( const UnboundedRegExpElement < SymbolType > & structure ) : UnboundedRegExpStructure ( std::move_copy ( structure ) ) {
}

template < class SymbolType >
UnboundedRegExpStructure < SymbolType >::UnboundedRegExpStructure ( ) : UnboundedRegExpStructure < SymbolType > ( UnboundedRegExpEmpty < SymbolType > ( ) ) {
}

template < class SymbolType >
UnboundedRegExpStructure < SymbolType >::UnboundedRegExpStructure ( const FormalRegExpStructure < SymbolType > & other ) : UnboundedRegExpStructure ( * other.getStructure ( ).asUnbounded ( ) ) {
}

template < class SymbolType >
const UnboundedRegExpElement < SymbolType > & UnboundedRegExpStructure < SymbolType >::getStructure ( ) const {
	return * m_structure;
}

template < class SymbolType >
UnboundedRegExpElement < SymbolType > & UnboundedRegExpStructure < SymbolType >::getStructure ( ) {
	return * m_structure;
}

template < class SymbolType >
void UnboundedRegExpStructure < SymbolType >::setStructure ( const UnboundedRegExpElement < SymbolType > & structure ) {
	setStructure ( std::move_copy ( structure ) );
}

template < class SymbolType >
void UnboundedRegExpStructure < SymbolType >::setStructure ( UnboundedRegExpElement < SymbolType > && param ) {
	this->m_structure = std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( std::move ( param ).plunder ( ) );
}

} /* namespace regexp */

namespace alib {

template < class SymbolType >
regexp::UnboundedRegExpStructure < SymbolType > xmlApi < regexp::UnboundedRegExpStructure < SymbolType > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return regexp::UnboundedRegExpStructure < SymbolType > ( regexp::RegExpFromXMLParser::parseUnboundedRegExpElement < SymbolType > ( input ) );
}

template < class SymbolType >
bool xmlApi < regexp::UnboundedRegExpStructure < SymbolType > >::first ( const std::deque < sax::Token >::const_iterator & ) {
	throw exception::CommonException ( "Unimplemented" );
}

template < class SymbolType >
void xmlApi < regexp::UnboundedRegExpStructure < SymbolType > >::compose ( std::deque < sax::Token > & output, const regexp::UnboundedRegExpStructure < SymbolType > & data ) {
	data.getStructure ( ).template accept < void, regexp::RegExpToXMLComposer::Unbounded > ( output );
}

} /* namespace alib */

#endif /* UNBOUNDED_REG_EXP_STRUCTURE_H_ */
