/*
 * FormalRTESymbolAlphabet.h
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_SYMBOL_ALPHABET_H_
#define FORMAL_RTE_SYMBOL_ALPHABET_H_

#include "FormalRTESymbol.h"

#include <exception/CommonException.h>
#include <sstream>

namespace rte {

/**
 * Represents symbol in the regular expression. Contains name of the symbol.
 */
template < class SymbolType, class RankType >
class FormalRTESymbolAlphabet : public FormalRTESymbol < SymbolType, RankType >, public std::VararyNode < std::smart_ptr < FormalRTESymbol < SymbolType, RankType > >, std::smart_ptr < const FormalRTESymbol < SymbolType, RankType > >, FormalRTESymbolAlphabet < SymbolType, RankType > > {
public:
	void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const {
		visitor.visit ( * this );
	}

	explicit FormalRTESymbolAlphabet ( std::ranked_symbol < SymbolType, RankType > symbol );

	/**
	 * @copydoc FormalRTEElement::clone() const
	 */
	virtual FormalRTESymbolAlphabet < SymbolType, RankType > * clone ( ) const;

	/**
	 * @copydoc FormalRTEElement::plunder() const
	 */
	virtual FormalRTESymbolAlphabet < SymbolType, RankType > * plunder ( ) &&;

	/**
	 * @copydoc FormalRTEElement::testSymbol() const
	 */
	virtual bool testSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) const;

	/**
	 * @copydoc FormalRTEElement::computeMinimalAlphabet()
	 */
	virtual void computeMinimalAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const;

	/**
	 * @copydoc FormalRTEElement::checkAlphabet()
	 */
	virtual bool checkAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const;

	const std::vector < std::smart_ptr < const FormalRTESymbol < SymbolType, RankType > > > & getElements ( ) const;
	const std::vector < std::smart_ptr < FormalRTESymbol < SymbolType, RankType > > > & getElements ( );

	void appendElement ( const FormalRTESymbol < SymbolType, RankType > & element );
	void appendElement ( FormalRTESymbol < SymbolType, RankType > && element );

	virtual int compare ( const FormalRTESymbolAlphabet & ) const;

	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	/**
	 * @copydoc FormalRTEElement::operator>>() const
	 */
	virtual void operator >>( std::ostream & out ) const;

	/**
	 * Returns the string representation of RTE Symbol.
	 */
	virtual explicit operator std::string ( ) const;
};

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType >::FormalRTESymbolAlphabet ( std::ranked_symbol < SymbolType, RankType > symbol ) : FormalRTESymbol < SymbolType, RankType > ( symbol ) {
}

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType > * FormalRTESymbolAlphabet < SymbolType, RankType >::clone ( ) const {
	return new FormalRTESymbolAlphabet ( * this );
}

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType > * FormalRTESymbolAlphabet < SymbolType, RankType >::plunder ( ) && {
	return new FormalRTESymbolAlphabet ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const std::vector < std::smart_ptr < const FormalRTESymbol < SymbolType, RankType > > > & FormalRTESymbolAlphabet < SymbolType, RankType >::getElements ( ) const {
	return this->getChildren ( );
}

template < class SymbolType, class RankType >
const std::vector < std::smart_ptr < FormalRTESymbol < SymbolType, RankType > > > & FormalRTESymbolAlphabet < SymbolType, RankType >::getElements ( ) {
	return this->getChildren ( );
}

template < class SymbolType, class RankType >
void FormalRTESymbolAlphabet < SymbolType, RankType >::appendElement ( const FormalRTESymbol < SymbolType, RankType > & element ) {
	return appendElement ( std::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESymbolAlphabet < SymbolType, RankType >::appendElement ( FormalRTESymbol < SymbolType, RankType > && element ) {
	// TODO: Remove this.  Pass vector in constructor. Now an Element can have LESS children that is MUST have.

	if ( this->getChildren ( ).size ( ) >= ( size_t ) this->getSymbol ( ).getRank ( ) )
		throw exception::CommonException ( "Ranked node cannot have more children then its rank" );

	this->pushBackChild ( std::smart_ptr < FormalRTESymbol < SymbolType, RankType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType, class RankType >
int FormalRTESymbolAlphabet < SymbolType, RankType >::compare ( const FormalRTESymbolAlphabet < SymbolType, RankType > & other ) const {
	auto first = std::tie ( this->getSymbol ( ), this->getChildren ( ) );
	auto second = std::tie ( other.getSymbol ( ), other.getChildren ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void FormalRTESymbolAlphabet < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTESymbolAlphabet " << " symbol = " << this->getSymbol ( ) << " children = " << this->getChildren ( ) << "})";
}

template < class SymbolType, class RankType >
bool FormalRTESymbolAlphabet < SymbolType, RankType >::testSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) const {
	return symbol == this->getSymbol ( );
}

template < class SymbolType, class RankType >
void FormalRTESymbolAlphabet < SymbolType, RankType >::computeMinimalAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	alphabetF.insert ( this->getSymbol ( ) );

	for ( const std::smart_ptr < const FormalRTESymbol < SymbolType, RankType > > & child : this->getChildren ( ) )
		child->computeMinimalAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
bool FormalRTESymbolAlphabet < SymbolType, RankType >::checkAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return alphabetF.count ( this->getSymbol ( ) ) > 0 && std::all_of ( getElements ( ).begin ( ), getElements ( ).end ( ), [&] ( const std::smart_ptr < const FormalRTESymbol < SymbolType, RankType > > & e ) {
			return e->checkAlphabet ( alphabetF, alphabetK );
		} );
}

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_SYMBOL_ALPHABET_H_ */
