/*
 * UnrankedPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRANKED_PATTERN_H_
#define UNRANKED_PATTERN_H_

#include <string>
#include <set>
#include <tree>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../TreeFeatures.h"
#include "../TreeException.h"
#include "../UnrankedTreeBase.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"

namespace tree {

class GeneralAlphabet;
class SubtreeWildcard;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
template < class SymbolType >
class UnrankedPattern : public UnrankedTreeBase, public std::Components < UnrankedPattern < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > {
	std::tree < SymbolType > content;

	void checkAlphabet ( const std::tree < SymbolType > & pattern ) const;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedPattern ( SymbolType subtreeWildcard, std::set < SymbolType > alphabet, std::tree < SymbolType > pattern );
	explicit UnrankedPattern ( SymbolType subtreeWildcard, std::tree < SymbolType > pattern );
	template < class RankType >
	explicit UnrankedPattern ( const RankedPattern < SymbolType, RankType > & other );

	/**
	 * @return Content of the pattern
	 */
	const std::tree < SymbolType > & getContent ( ) const;

	virtual const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const SymbolType & getSubtreeWildcard ( ) const {
		return this->template accessElement < SubtreeWildcard > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( std::tree < SymbolType > pattern );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedPattern";

		return xmlTagName;
	}

	static UnrankedPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

} /* namespace tree */

#include "../ranked/RankedPattern.h"

namespace tree {

template < class SymbolType >
UnrankedPattern < SymbolType >::UnrankedPattern ( SymbolType subtreeWildcard, std::set < SymbolType > alphabet, std::tree < SymbolType > pattern ) : std::Components < UnrankedPattern, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), content ( std::move ( pattern ) ) {
	checkAlphabet ( content );
}

template < class SymbolType >
UnrankedPattern < SymbolType >::UnrankedPattern ( SymbolType subtreeWildcard, std::tree < SymbolType > pattern ) : UnrankedPattern ( subtreeWildcard, std::set < SymbolType > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ) + std::set < SymbolType > { subtreeWildcard }, pattern ) {
}

template < class SymbolType >
template < class RankType >
UnrankedPattern < SymbolType >::UnrankedPattern ( const RankedPattern < SymbolType, RankType > & other ) : UnrankedPattern ( other.getSubtreeWildcard ( ).getSymbol ( ), TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), TreeAuxiliary::toUnrankedTree ( other.getContent ( ) ) ) {
}

template < class SymbolType >
UnrankedTreeBase * UnrankedPattern < SymbolType >::clone ( ) const {
	return new UnrankedPattern ( * this );
}

template < class SymbolType >
UnrankedTreeBase * UnrankedPattern < SymbolType >::plunder ( ) && {
	return new UnrankedPattern ( std::move ( * this ) );
}

template < class SymbolType >
const std::tree < SymbolType > & UnrankedPattern < SymbolType >::getContent ( ) const {
	return content;
}

template < class SymbolType >
void UnrankedPattern < SymbolType >::checkAlphabet ( const std::tree < SymbolType > & data ) const {
	std::set < SymbolType > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < SymbolType > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType >
void UnrankedPattern < SymbolType >::setTree ( std::tree < SymbolType > pattern ) {
	checkAlphabet ( pattern );

	this->content = std::move ( pattern );
}

template < class SymbolType >
void UnrankedPattern < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(UnrankedPattern " << this->content << ")";
}

template < class SymbolType >
int UnrankedPattern < SymbolType >::compare ( const UnrankedPattern & other ) const {
	auto first = std::tie ( content, getAlphabet(), getSubtreeWildcard() );
	auto second = std::tie ( other.content, other.getAlphabet(), other.getSubtreeWildcard() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void UnrankedPattern < SymbolType >::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

template < class SymbolType >
UnrankedPattern < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
UnrankedPattern < SymbolType > UnrankedPattern < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedPattern::getXmlTagName() );
	SymbolType subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol < SymbolType > ( input );
	std::set < SymbolType > rankedAlphabet = TreeFromXMLParser::parseAlphabet < SymbolType > ( input );
	std::tree < SymbolType > root = TreeFromXMLParser::parseTreeContent < SymbolType > ( input );
	UnrankedPattern < SymbolType > tree ( std::move ( subtreeWildcardSymbol ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedPattern::getXmlTagName() );
	return tree;
}

template < class SymbolType >
void UnrankedPattern < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnrankedPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( UnrankedPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* UnrankedPattern < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace std {

template < class SymbolType >
class ComponentConstraint< ::tree::UnrankedPattern < SymbolType >, SymbolType, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::UnrankedPattern < SymbolType > & pattern, const SymbolType & symbol ) {
		const std::tree<SymbolType>& content = pattern.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end() || pattern.template accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::UnrankedPattern < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::tree::UnrankedPattern < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
class ElementConstraint< ::tree::UnrankedPattern < SymbolType >, SymbolType, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::UnrankedPattern < SymbolType > & pattern, const SymbolType & symbol ) {
		return pattern.template accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::UnrankedPattern < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* UNRANKED_PATTERN_H_ */
