#include "AutomatonTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "automaton/FSM/DFA.h"
#include "automaton/FSM/ExtendedNFA.h"
#include "automaton/PDA/SinglePopDPDA.h"
#include "automaton/PDA/DPDA.h"
#include "automaton/PDA/NPDA.h"
#include "automaton/PDA/RealTimeHeightDeterministicNPDA.h"
#include "automaton/TA/NFTA.h"

#include "automaton/AutomatonException.h"

#include "factory/XmlDataFactory.hpp"

#include "alphabet/LabeledSymbol.h"
#include "alphabet/BottomOfTheStackSymbol.h"

#include "regexp/RegExp.h"
#include "regexp/unbounded/UnboundedRegExp.h"
#include "regexp/unbounded/UnboundedRegExpElements.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AutomatonTest, "automaton" );
CPPUNIT_TEST_SUITE_REGISTRATION( AutomatonTest );

void AutomatonTest::setUp() {
}

void AutomatonTest::tearDown() {
}

void AutomatonTest::testDFAParser() {
	automaton::DFA<> automaton(DefaultStateType(1));

	automaton.addState(DefaultStateType(1));
	automaton.addState(DefaultStateType(2));
	automaton.addState(DefaultStateType(3));
	automaton.addInputSymbol(DefaultSymbolType("a"));
	automaton.addInputSymbol(DefaultSymbolType("b"));

	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a"), DefaultStateType(2));
	automaton.addTransition(DefaultStateType(2), DefaultSymbolType("b"), DefaultStateType(1));

	automaton.addFinalState(DefaultStateType(3));

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::DFA<> automaton2 = alib::XmlDataFactory::fromTokens<automaton::DFA<>>(std::move(tokens2));

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::SinglePopDPDATransitions() {
	automaton::SinglePopDPDA < > automaton(DefaultStateType(1), DefaultSymbolType("S"));

	automaton.addState(DefaultStateType(1));
	automaton.addState(DefaultStateType(2));
	automaton.addState(DefaultStateType(3));

	automaton.addInputSymbol(DefaultSymbolType("a"));
	automaton.addInputSymbol(DefaultSymbolType("b"));

	automaton.addPushdownStoreSymbol(DefaultSymbolType("S"));
	automaton.addPushdownStoreSymbol(DefaultSymbolType("X"));
	automaton.addPushdownStoreSymbol(DefaultSymbolType("Y"));

	automaton.addTransition(DefaultStateType(3), /* eps, */ DefaultSymbolType("X"), DefaultStateType(1), { DefaultSymbolType("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(DefaultStateType(3), /* eps, */ DefaultSymbolType("X"), DefaultStateType(1), { DefaultSymbolType("X") } ) );

	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a"), DefaultSymbolType("Y"), DefaultStateType(1), { DefaultSymbolType("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a"), DefaultSymbolType("X"), DefaultStateType(1), { DefaultSymbolType("X") } ),
		automaton::AutomatonException
	);

	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a"), DefaultSymbolType("X"), DefaultStateType(2), { DefaultSymbolType("Y") } );
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a"), DefaultSymbolType("X"), DefaultStateType(2), { DefaultSymbolType("Y") } );
	automaton.addTransition(DefaultStateType(1), /* eps, */ DefaultSymbolType("Y"), DefaultStateType(2), { DefaultSymbolType("Y") } );

	automaton.addTransition(DefaultStateType(2), DefaultSymbolType("b"), DefaultSymbolType("X"), DefaultStateType(1), { DefaultSymbolType("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(DefaultStateType(2), /* eps, */ DefaultSymbolType("X"), DefaultStateType(1), { DefaultSymbolType("X") } ),
		automaton::AutomatonException
	);

	automaton.addFinalState(DefaultStateType(3));
}

void AutomatonTest::DPDATransitions() {
	automaton::DPDA < > automaton(DefaultStateType(1), DefaultSymbolType("S"));

	automaton.addState(DefaultStateType(1));
	automaton.addState(DefaultStateType(2));
	automaton.addState(DefaultStateType(3));
	automaton.addState(DefaultStateType(4));

	automaton.addInputSymbol(DefaultSymbolType("a"));
	automaton.addInputSymbol(DefaultSymbolType("b"));

	automaton.addPushdownStoreSymbol(DefaultSymbolType("S"));
	automaton.addPushdownStoreSymbol(DefaultSymbolType("X"));
	automaton.addPushdownStoreSymbol(DefaultSymbolType("Y"));

	automaton.addTransition(DefaultStateType(3), /* eps, */ { DefaultSymbolType("X") }, DefaultStateType(1), { DefaultSymbolType("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(DefaultStateType(3), /* eps, */ { DefaultSymbolType("X") }, DefaultStateType(1), { DefaultSymbolType("X") } ) );

	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a"), { DefaultSymbolType("Y") }, DefaultStateType(1), { DefaultSymbolType("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a"), { DefaultSymbolType("X") }, DefaultStateType(1), { DefaultSymbolType("X") } ),
		automaton::AutomatonException
	);

	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a"), { DefaultSymbolType("X") }, DefaultStateType(2), { DefaultSymbolType("Y") } );
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a"), { DefaultSymbolType("X") }, DefaultStateType(2), { DefaultSymbolType("Y") } );
	automaton.addTransition(DefaultStateType(1), /* eps, */ { DefaultSymbolType("Y") }, DefaultStateType(2), { DefaultSymbolType("Y") } );

	automaton.addTransition(DefaultStateType(2), DefaultSymbolType("b"), { DefaultSymbolType("X") }, DefaultStateType(1), { DefaultSymbolType("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(DefaultStateType(2), /* eps, */ { DefaultSymbolType("X") }, DefaultStateType(1), { DefaultSymbolType("X") } ),
		automaton::AutomatonException
	);

	automaton.addTransition(DefaultStateType(4), DefaultSymbolType("a"), { DefaultSymbolType("X"), DefaultSymbolType("Y") }, DefaultStateType(2), { DefaultSymbolType("Y") } );
	automaton.addTransition(DefaultStateType(4), DefaultSymbolType("a"), { DefaultSymbolType("X"), DefaultSymbolType("S") }, DefaultStateType(2), { DefaultSymbolType("Y") } );
	automaton.addTransition(DefaultStateType(4), /* eps, */ { DefaultSymbolType("X"), DefaultSymbolType("X") }, DefaultStateType(2), { DefaultSymbolType("Y") } );
	CPPUNIT_ASSERT(!automaton.addTransition(DefaultStateType(4), /* eps, */ { DefaultSymbolType("X"), DefaultSymbolType("X") }, DefaultStateType(2), { DefaultSymbolType("Y") } ));
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(DefaultStateType(4), /* eps, */ { DefaultSymbolType("X"), DefaultSymbolType("Y") }, DefaultStateType(2), { DefaultSymbolType("Y") } ),
		automaton::AutomatonException
	);
	automaton.addTransition(DefaultStateType(4), DefaultSymbolType("b"), { DefaultSymbolType("X"), DefaultSymbolType("Y") }, DefaultStateType(2), { DefaultSymbolType("Y") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(DefaultStateType(4), DefaultSymbolType("a"), { DefaultSymbolType("X") }, DefaultStateType(2), { DefaultSymbolType("Y") } ),
		automaton::AutomatonException
	);

	automaton.addFinalState(DefaultStateType(3));
}

void AutomatonTest::testExtendedNFAAlphabet() {
	automaton::ExtendedNFA < > automaton (DefaultStateType(0));
	DefaultStateType s0 = DefaultStateType(0);
	DefaultStateType s1 = DefaultStateType(1);
	automaton.addState(s0);
	automaton.addState(s1);

	automaton.setInputAlphabet({DefaultSymbolType('a'), DefaultSymbolType('b') });

	regexp::UnboundedRegExpConcatenation < DefaultSymbolType > goodConcat1, goodConcat2;
	regexp::UnboundedRegExpConcatenation < DefaultSymbolType > badConcat;

	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol < DefaultSymbolType > (DefaultSymbolType('a')));
	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol < DefaultSymbolType > (DefaultSymbolType('b')));

	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol < DefaultSymbolType > (DefaultSymbolType('b')));
	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol < DefaultSymbolType > (DefaultSymbolType('b')));

	badConcat.appendElement(regexp::UnboundedRegExpSymbol < DefaultSymbolType > (DefaultSymbolType('b')));
	badConcat.appendElement(regexp::UnboundedRegExpSymbol < DefaultSymbolType > (DefaultSymbolType('d')));

	CPPUNIT_ASSERT_THROW(automaton.addTransition(s0, regexp::UnboundedRegExpStructure < DefaultSymbolType > (badConcat), s1), automaton::AutomatonException);
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::UnboundedRegExpStructure < DefaultSymbolType > (goodConcat1), s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::UnboundedRegExpStructure < DefaultSymbolType > (goodConcat2), s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::UnboundedRegExpStructure < DefaultSymbolType > (regexp::UnboundedRegExpEpsilon < DefaultSymbolType > ()), s1));
}

void AutomatonTest::testNPDATransitions() {
	automaton::NPDA < > automaton(DefaultStateType(0), DefaultSymbolType("T"));

	automaton.setStates({DefaultStateType(0), DefaultStateType(1), DefaultStateType(2), DefaultStateType(3), DefaultStateType(4)});
	automaton.setInputAlphabet({DefaultSymbolType("a2"), DefaultSymbolType("a1"), DefaultSymbolType("a0")});
	automaton.setPushdownStoreAlphabet({DefaultSymbolType("T"), DefaultSymbolType("R")});

	automaton.addTransition(DefaultStateType(0), DefaultSymbolType("a2"), {DefaultSymbolType("T")}, DefaultStateType(0), {DefaultSymbolType("T"), DefaultSymbolType("T")});
	automaton.addTransition(DefaultStateType(0), DefaultSymbolType("a1"), {DefaultSymbolType("T")}, DefaultStateType(0), {DefaultSymbolType("T")});
	automaton.addTransition(DefaultStateType(0), DefaultSymbolType("a0"), {DefaultSymbolType("T")}, DefaultStateType(0), {});

	automaton.addTransition(DefaultStateType(0), DefaultSymbolType("a2"), {DefaultSymbolType("T")}, DefaultStateType(1), {DefaultSymbolType("R"), DefaultSymbolType("T")});

	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a2"), {DefaultSymbolType("T")}, DefaultStateType(1), {DefaultSymbolType("T"), DefaultSymbolType("T")});
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a2"), {DefaultSymbolType("R")}, DefaultStateType(1), {DefaultSymbolType("T"), DefaultSymbolType("R")});
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a1"), {DefaultSymbolType("T")}, DefaultStateType(1), {DefaultSymbolType("T")});
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a1"), {DefaultSymbolType("R")}, DefaultStateType(1), {DefaultSymbolType("R")});
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a0"), {DefaultSymbolType("T")}, DefaultStateType(1), {});
	automaton.addTransition(DefaultStateType(1), DefaultSymbolType("a0"), {DefaultSymbolType("R")}, DefaultStateType(2), {});

	automaton.addTransition(DefaultStateType(2), DefaultSymbolType("a1"), {DefaultSymbolType("T")}, DefaultStateType(3), {DefaultSymbolType("R")});

	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a2"), {DefaultSymbolType("T")}, DefaultStateType(3), {DefaultSymbolType("T"), DefaultSymbolType("T")});
	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a2"), {DefaultSymbolType("R")}, DefaultStateType(3), {DefaultSymbolType("T"), DefaultSymbolType("R")});
	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a1"), {DefaultSymbolType("T")}, DefaultStateType(3), {DefaultSymbolType("T")});
	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a1"), {DefaultSymbolType("R")}, DefaultStateType(3), {DefaultSymbolType("R")});
	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a0"), {DefaultSymbolType("T")}, DefaultStateType(3), {});
	automaton.addTransition(DefaultStateType(3), DefaultSymbolType("a0"), {DefaultSymbolType("R")}, DefaultStateType(4), {});

	automaton.setFinalStates({DefaultStateType(4)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testRHPDATransitions() {
	automaton::RealTimeHeightDeterministicNPDA < > automaton{alphabet::BottomOfTheStackSymbol::instance < DefaultSymbolType > ( ) };
	automaton.setStates({DefaultStateType(1), DefaultStateType(2), DefaultStateType(3)});
	automaton.setInitialStates({DefaultStateType(1)});

	automaton.setInputAlphabet({DefaultSymbolType("a"), DefaultSymbolType("b")});
	automaton.setPushdownStoreAlphabet({alphabet::BottomOfTheStackSymbol::instance < DefaultSymbolType > ( ), DefaultSymbolType("X"), DefaultSymbolType("Y")});

	automaton.addCallTransition(DefaultStateType(1), DefaultSymbolType("a"), DefaultStateType(2), DefaultSymbolType("X"));
	automaton.addCallTransition(DefaultStateType(2), DefaultStateType(3), DefaultSymbolType("X"));
	automaton.addReturnTransition(DefaultStateType(3), DefaultSymbolType("Y"), DefaultStateType(1));

	automaton.setFinalStates({DefaultStateType(3)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testNFTAParser() {
	std::ranked_symbol < > a ('a', 2);
	std::ranked_symbol < > b ('b', 1);
	std::ranked_symbol < > c ('c', 0);

	DefaultStateType q2 = DefaultStateType ("q2");
	DefaultStateType q1 = DefaultStateType ("q1");
	DefaultStateType q0 = DefaultStateType ("q0");

	automaton::NFTA < > automaton;

	automaton.addState(q2);
	automaton.addState(q1);
	automaton.addState(q0);
	automaton.addInputSymbol(a);
	automaton.addInputSymbol(b);
	automaton.addInputSymbol(c);

	std::vector<DefaultStateType> states1 = {q1, q0};
	automaton.addTransition(a, states1, q2);
	automaton.addTransition(a, states1, q1);
	std::vector<DefaultStateType> states2 = {q0};
	automaton.addTransition(b, states2, q1);
	automaton.addTransition(b, states2, q0);
	std::vector<DefaultStateType> states3 = {};
	automaton.addTransition(c, states3, q0);

	automaton.addFinalState(q2);

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);
		std::cout << tmp << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::NFTA < > automaton2 = alib::XmlDataFactory::fromTokens<automaton::NFTA < > >(std::move(tokens2));

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::testNFTATransitions() {
	std::ranked_symbol < > a ('a', 2);
	std::ranked_symbol < > b ('b', 1);
	std::ranked_symbol < > c ('c', 0);
	std::ranked_symbol < > d ('d', 0);

	DefaultStateType q2 = DefaultStateType ("q2");
	DefaultStateType q1 = DefaultStateType ("q1");
	DefaultStateType q0 = DefaultStateType ("q0");

	automaton::NFTA < > automaton;

	automaton.addState(q2);
	automaton.addState(q1);
	automaton.addState(q0);
	automaton.addInputSymbol(a);
	automaton.addInputSymbol(b);
	automaton.addInputSymbol(c);

	std::vector<DefaultStateType> states1 = {q1, q0};
	automaton.addTransition(a, states1, q2);
	automaton.addTransition(a, states1, q1);
	std::vector<DefaultStateType> states2 = {q0};
	automaton.addTransition(b, states2, q1);
	automaton.addTransition(b, states2, q0);
	std::vector<DefaultStateType> states3 = {};
	automaton.addTransition(c, states3, q0);

	automaton.addFinalState(q2);

	CPPUNIT_ASSERT_THROW(automaton.removeInputSymbol(a), exception::CommonException);
	CPPUNIT_ASSERT_NO_THROW(automaton.removeInputSymbol(d));
	CPPUNIT_ASSERT_NO_THROW(automaton.removeTransition(b, states1, q2));
	CPPUNIT_ASSERT_THROW(automaton.addTransition(b, states1, q1), exception::CommonException);
	CPPUNIT_ASSERT_THROW(automaton.addTransition(a, states2, q1), exception::CommonException);
	CPPUNIT_ASSERT_THROW(automaton.addTransition(d, states3, q0), exception::CommonException);
}
