/*
 * AdjacencyMatrixDirectedGraph.h
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#ifndef ADJACENCY_MATRIX_DIRECTED_GRAPH_H_
#define ADJACENCY_MATRIX_DIRECTED_GRAPH_H_

#include <set>
#include <vector>
#include <unordered_map>

#include "DirectedGraph.h"

namespace graph {

class UndirectedGraph;

/*
 *  AdjacencyMatrix representation
 *
 *  unordered_map:
 *    [ -- ][ n1 ][ n2 ][ n3 ][ n4 ]
 *    [ n1 ]  0     2     0     1
 *    [ n2 ]  1     0     0     1
 *    [ n3 ]  0     0     0     1
 *    [ n4 ]  1     0     1     0
 *
 *  edges:
 *    n1 -> n2, n1 -> n2, n1 -> n4
 *    n2 -> n1, n2 -> n4
 *    n3 -> n4
 *    n4 -> n1, n4 -> n3
 *
 */

class AdjacencyMatrixDirectedGraph : public DirectedGraph {
	std::unordered_map<Node, std::unordered_map<Node, int>> adj;
	std::set<DirectedEdge> edges;

public:
	AdjacencyMatrixDirectedGraph();

	/// constructs an directed adjacency matrix from any representation of directed graph
	AdjacencyMatrixDirectedGraph( const DirectedGraph & graph );
	/// converts undirected graph into directed by making 2 directed edges in both directions from each undirected one
	AdjacencyMatrixDirectedGraph( const UndirectedGraph & ugraph );

	virtual GraphBase * clone() const override;
	virtual GraphBase * plunder() && override;

	virtual REPRESENTATION getRepresentation() const override;
	virtual std::string className() const override;

	virtual std::set< Node > getNodes() const override;
	virtual std::set< DirectedEdge > getEdges() const override;

	virtual size_t nodeCount() const override;
	virtual size_t edgeCount() const override;

	virtual bool hasNode( const Node & node ) const override;
	virtual bool hasEdge( const DirectedEdge & edge ) const override;

	virtual bool hasEdge( const Node & from, const Node & to ) const override;
	virtual std::set< DirectedEdge > findEdges( const Node & from, const Node & to ) const override;

	virtual std::set< Node > neighbors( const Node & node ) const override;
	virtual std::set< DirectedEdge > neighborEdges( const Node & node ) const override;

	virtual std::set< Node > successors( const Node & node ) const override;
	virtual std::set< Node > predecessors( const Node & node ) const override;
	virtual std::set< DirectedEdge > successorEdges( const Node & node ) const override;
	virtual std::set< DirectedEdge > predecessorEdges( const Node & node ) const override;

	virtual bool addNode( const Node & node) override;
	virtual bool addNode( const Node & node, int value ) override;
	virtual bool removeNode( const Node & node) override;

	virtual bool addEdge( const DirectedEdge & edge ) override;
	virtual bool addEdge( const DirectedEdge & edge, int value ) override;
	virtual bool removeEdge( const DirectedEdge & edge ) override;

	static AdjacencyMatrixDirectedGraph parse( std::deque< sax::Token >::iterator & input );
	virtual void compose( std::deque< sax::Token > & out ) const override;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "AdjacencyMatrixDirectedGraph";

		return xmlTagName;
	}

	alib::ObjectBase * inc ( ) && override;
};

} // namespace graph

#endif // ADJACENCY_MATRIX_DIRECTED_GRAPH_H_
