/*
 * ReachableStates.h
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef EFFICIENT_REACHABLE_STATES_H_
#define EFFICIENT_REACHABLE_STATES_H_

#include <core/multipleDispatch.hpp>
#include <algorithm>
#include <deque>
#include <set>
#include <map>
#include <queue>

#include <automaton/Automaton.h>
#include <automaton/AutomatonFeatures.h>

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

namespace automaton {

namespace properties {

namespace efficient {

class ReachableStates : public std::SingleDispatch<ReachableStates, std::set<DefaultStateType>, const automaton::AutomatonBase &> {
public:
	static std::set<DefaultStateType> reachableStates( const automaton::Automaton & automaton );

	/**
	 * Removes dead states from FSM. Melichar 2.29
	 */
	template < class T, class SymbolType = typename automaton::SymbolTypeOfAutomaton < T >, class StateType = typename automaton::StateTypeOfAutomaton < T > >
	static std::set<StateType> reachableStates( const T & fsm );
	template < class SymbolType, class StateType >
	static std::set<StateType> reachableStates( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm );
	template < class SymbolType, class StateType >
	static std::set<StateType> reachableStates( const automaton::DFA < SymbolType, StateType > & fsm );
};

template < class T, class SymbolType, class StateType >
std::set<StateType> ReachableStates::reachableStates( const T & fsm ) {
	std::map<StateType, std::set<StateType>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	std::deque<StateType> queue { fsm.getInitialState( ) };
	std::set<StateType> visited { fsm.getInitialState( ) };

	while( !queue.empty() ) {
		const std::set<StateType>& to = transitions[queue.front()];
		queue.pop_front();

		for(const StateType& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<StateType&>(process)));
			}
	}

	return visited;
}

template < class SymbolType, class StateType >
std::set<StateType> ReachableStates::reachableStates( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm ) {
	std::map<StateType, std::set<StateType>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	std::deque<StateType> queue ( fsm.getInitialStates( ).begin(), fsm.getInitialStates().end() );
	std::set<StateType> visited = fsm.getInitialStates( );

	while( !queue.empty() ) {
		const std::set<StateType>& to = transitions[queue.front()];
		queue.pop_front();

		for(const StateType& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<StateType&>(process)));
			}
	}

	return visited;
}

template < class SymbolType, class StateType >
std::set<StateType> ReachableStates::reachableStates( const automaton::DFA < SymbolType, StateType > & fsm ) {
	std::map<StateType, std::set<StateType>> transitions;
	for(const auto& transition : fsm.getTransitions())
		transitions[transition.first.first].insert(transition.second);

	std::deque<StateType> queue { fsm.getInitialState( ) };
	std::set<StateType> visited { fsm.getInitialState( ) };

	while( !queue.empty() ) {
		const std::set<StateType>& to = transitions[queue.front()];
		queue.pop_front();

		for(const StateType& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<StateType&>(process)));
			}
	}

	return visited;
}

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

#endif /* EFFICIENT_REACHABLE_STATES_H_ */
