#ifndef GRAMMAR_TO_STRING_COMPOSER_H_
#define GRAMMAR_TO_STRING_COMPOSER_H_

#include <ostream>
#include <core/multipleDispatch.hpp>
#include "grammar/Grammar.h"
#include "grammar/GrammarFeatures.h"

#include <object/ObjectToStringComposer.h>

namespace grammar {

class GrammarToStringComposer: public std::SingleDispatchFirstStaticParam<GrammarToStringComposer, void, std::ostream&, const GrammarBase &> {
public:
	template<class T>
	static void composeCFLikeGrammar(std::ostream& output, const T& grammar);
	template<class T>
	static void composeCSLikeGrammar(std::ostream& output, const T& grammar);
	template<class T>
	static void composePreservingCSLikeGrammar(std::ostream& output, const T& grammar);

	/**
	 * Prints XML representation of Automaton to the output stream.
	 * @param automaton automaton to print
	 * @return list of xml tokens representing the automaton
	 */
	static void compose(std::ostream& output, const Grammar& grammar);

	template<class SymbolType>
	static void compose(std::ostream& output, const LeftLG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const LeftRG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const RightLG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const RightRG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const LG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const CFG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const EpsilonFreeCFG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const CNF < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const GNF < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const CSG < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const NonContractingGrammar < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const ContextPreservingUnrestrictedGrammar < SymbolType > & grammar);
	template<class SymbolType>
	static void compose(std::ostream& output, const UnrestrictedGrammar < SymbolType > & grammar);
};

template<class T>
void GrammarToStringComposer::composeCFLikeGrammar(std::ostream& output, const T& grammar) {
	bool first;

	output << " (" << std::endl;

	output << "{";
	first = false;
	for(const auto& symbol : grammar.getNonterminalAlphabet() ) {
		if(first)
			output << ", ";
		else
			first = true;
		alib::ObjectToStringComposer::compose(output, symbol);
	}
	output << "}," << std::endl;
	output << "{";
	first = false;
	for(const auto& symbol : grammar.getTerminalAlphabet() ) {
		if(first)
			output << ", ";
		else
			first = true;
		alib::ObjectToStringComposer::compose(output, symbol);
	}
	output << "}," << std::endl;
	output << "{ ";
	first = true;
	for(const auto& rule : grammar.getRawRules() ) {
		if(first)
			first = false;
		else
			output << "," << std::endl << "  ";
		alib::ObjectToStringComposer::compose(output, rule.first);
		output << " ->";
		bool innerFirst = true;
		for(const auto& rhs : rule.second) {
			if(innerFirst)
				innerFirst = false;
			else
				output << " |";
			for(const auto& symbol : rhs) {
				output << " ";
				alib::ObjectToStringComposer::compose(output, symbol);
			}
		}
	}
	output << "}," << std::endl;
	alib::ObjectToStringComposer::compose(output, grammar.getInitialSymbol());
	output << ")" << std::endl;
}

template<class T>
void GrammarToStringComposer::composeCSLikeGrammar(std::ostream& output, const T& grammar) {
	bool first;

	output << " (" << std::endl;

	output << "{";
	first = false;
	for(const auto& symbol : grammar.getNonterminalAlphabet() ) {
		if(first)
			output << ", ";
		else
			first = true;
		alib::ObjectToStringComposer::compose(output, symbol);
	}
	output << "}," << std::endl;
	output << "{";
	first = false;
	for(const auto& symbol : grammar.getTerminalAlphabet() ) {
		if(first)
			output << ", ";
		else
			first = true;
		alib::ObjectToStringComposer::compose(output, symbol);
	}
	output << "}," << std::endl;
	output << "{";
	first = true;
	for(const auto& rule : grammar.getRules() ) {
		if(first)
			first = false;
		else
			output << "," << std::endl << " ";
		for(const auto& symbol : rule.first) {
			output << " ";
			alib::ObjectToStringComposer::compose(output, symbol);
		}
		output << " ->";
		bool innerFirst = true;
		for(const auto& rhs : rule.second) {
			if(innerFirst)
				innerFirst = false;
			else
				output << " |";
			for(const auto& symbol : rhs) {
				output << " ";
				alib::ObjectToStringComposer::compose(output, symbol);
			}
		}
	}
	output << "}," << std::endl;
	alib::ObjectToStringComposer::compose(output, grammar.getInitialSymbol());
	output << ")" << std::endl;
}

template<class T>
void GrammarToStringComposer::composePreservingCSLikeGrammar(std::ostream& output, const T& grammar) {
	bool first;

	output << " (" << std::endl;

	output << "{";
	first = false;
	for(const auto& symbol : grammar.getNonterminalAlphabet() ) {
		if(first)
			output << ", ";
		else
			first = true;
		alib::ObjectToStringComposer::compose(output, symbol);
	}
	output << "}," << std::endl;
	output << "{";
	first = false;
	for(const auto& symbol : grammar.getTerminalAlphabet() ) {
		if(first)
			output << ", ";
		else
			first = true;
		alib::ObjectToStringComposer::compose(output, symbol);
	}
	output << "}," << std::endl;
	output << "{";
	first = true;
	for(const auto& rule : grammar.getRules() ) {
		if(first)
			first = false;
		else
			output << "," << std::endl << " ";
		for(const auto& symbol : std::get<0>(rule.first)) {
			output << " ";
			alib::ObjectToStringComposer::compose(output, symbol);
		}
		output << " | ";
		alib::ObjectToStringComposer::compose(output, std::get<1>(rule.first));
		output << " |";
		for(const auto& symbol : std::get<2>(rule.first)) {
			output << " ";
			alib::ObjectToStringComposer::compose(output, symbol);
		}
		output << " ->";
		bool innerFirst = true;
		for(const auto& rhs : rule.second) {
			if(innerFirst)
				innerFirst = false;
			else
				output << " |";
			for(const auto& symbol : rhs) {
				output << " ";
				alib::ObjectToStringComposer::compose(output, symbol);
			}
		}
	}
	output << "}," << std::endl;
	alib::ObjectToStringComposer::compose(output, grammar.getInitialSymbol());
	output << ")" << std::endl;
}

} /* namespace grammar */

#endif /* GRAMMAR_TO_STRING_COMPOSER_H_ */
